# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest


def test_detection3dpc(detection3dpc) -> None:
    # def test_oriented_bounding_box(detection3dpc):
    """Test oriented bounding box calculation and values."""
    obb = detection3dpc.get_oriented_bounding_box()
    assert obb is not None, "Oriented bounding box should not be None"

    # Verify OBB center values
    assert obb.center[0] == pytest.approx(-3.36002, abs=0.1)
    assert obb.center[1] == pytest.approx(-0.196446, abs=0.1)
    assert obb.center[2] == pytest.approx(0.220184, abs=0.1)

    # Verify OBB extent values
    assert obb.extent[0] == pytest.approx(0.531275, abs=0.12)
    assert obb.extent[1] == pytest.approx(0.461054, abs=0.1)
    assert obb.extent[2] == pytest.approx(0.155, abs=0.1)

    # def test_bounding_box_dimensions(detection3dpc):
    """Test bounding box dimension calculation."""
    dims = detection3dpc.get_bounding_box_dimensions()
    assert len(dims) == 3, "Bounding box dimensions should have 3 values"
    assert dims[0] == pytest.approx(0.350, abs=0.1)
    assert dims[1] == pytest.approx(0.250, abs=0.1)
    assert dims[2] == pytest.approx(0.550, abs=0.1)

    # def test_axis_aligned_bounding_box(detection3dpc):
    """Test axis-aligned bounding box calculation."""
    aabb = detection3dpc.get_bounding_box()
    assert aabb is not None, "Axis-aligned bounding box should not be None"

    # Verify AABB min values
    assert aabb.min_bound[0] == pytest.approx(-3.575, abs=0.1)
    assert aabb.min_bound[1] == pytest.approx(-0.375, abs=0.1)
    assert aabb.min_bound[2] == pytest.approx(-0.075, abs=0.1)

    # Verify AABB max values
    assert aabb.max_bound[0] == pytest.approx(-3.075, abs=0.1)
    assert aabb.max_bound[1] == pytest.approx(-0.125, abs=0.1)
    assert aabb.max_bound[2] == pytest.approx(0.475, abs=0.1)

    # def test_point_cloud_properties(detection3dpc):
    """Test point cloud data and boundaries."""
    points, _ = detection3dpc.pointcloud.as_numpy()
    assert len(points) > 60
    assert detection3dpc.pointcloud.frame_id == "world", (
        f"Expected frame_id 'world', got '{detection3dpc.pointcloud.frame_id}'"
    )

    min_pt = np.min(points, axis=0)
    max_pt = np.max(points, axis=0)
    center = np.mean(points, axis=0)

    # Verify point cloud boundaries
    assert min_pt[0] == pytest.approx(-3.575, abs=0.1)
    assert min_pt[1] == pytest.approx(-0.375, abs=0.1)
    assert min_pt[2] == pytest.approx(-0.075, abs=0.1)

    assert max_pt[0] == pytest.approx(-3.075, abs=0.1)
    assert max_pt[1] == pytest.approx(-0.125, abs=0.1)
    assert max_pt[2] == pytest.approx(0.475, abs=0.1)

    assert center[0] == pytest.approx(-3.326, abs=0.1)
    assert center[1] == pytest.approx(-0.202, abs=0.1)
    assert center[2] == pytest.approx(0.160, abs=0.1)

    # def test_foxglove_scene_entity_generation(detection3dpc):
    """Test Foxglove scene entity creation and structure."""
    entity = detection3dpc.to_foxglove_scene_entity("test_entity_123")

    # Verify entity metadata
    assert entity.id == "1", f"Expected entity ID '1', got '{entity.id}'"
    assert entity.frame_id == "world", f"Expected frame_id 'world', got '{entity.frame_id}'"
    assert entity.cubes_length == 1, f"Expected 1 cube, got {entity.cubes_length}"
    assert entity.texts_length == 1, f"Expected 1 text, got {entity.texts_length}"

    # def test_foxglove_cube_properties(detection3dpc):
    """Test Foxglove cube primitive properties."""
    entity = detection3dpc.to_foxglove_scene_entity("test_entity_123")
    cube = entity.cubes[0]

    # Verify position
    assert cube.pose.position.x == pytest.approx(-3.325, abs=0.1)
    assert cube.pose.position.y == pytest.approx(-0.250, abs=0.1)
    assert cube.pose.position.z == pytest.approx(0.200, abs=0.1)

    # Verify size
    assert cube.size.x == pytest.approx(0.350, abs=0.1)
    assert cube.size.y == pytest.approx(0.250, abs=0.1)
    assert cube.size.z == pytest.approx(0.550, abs=0.1)

    # Verify color (green with alpha)
    assert cube.color.r == pytest.approx(0.08235294117647059, abs=0.1)
    assert cube.color.g == pytest.approx(0.7176470588235294, abs=0.1)
    assert cube.color.b == pytest.approx(0.28627450980392155, abs=0.1)
    assert cube.color.a == pytest.approx(0.2, abs=0.1)

    # def test_foxglove_text_label(detection3dpc):
    """Test Foxglove text label properties."""
    entity = detection3dpc.to_foxglove_scene_entity("test_entity_123")
    text = entity.texts[0]

    assert text.text in ["1/suitcase (81%)", "1/suitcase (82%)"], (
        f"Expected text '1/suitcase (81%)' or '1/suitcase (82%)', got '{text.text}'"
    )
    assert text.pose.position.x == pytest.approx(-3.325, abs=0.1)
    assert text.pose.position.y == pytest.approx(-0.250, abs=0.1)
    assert text.pose.position.z == pytest.approx(0.575, abs=0.1)
    assert text.font_size == 20.0, f"Expected font size 20.0, got {text.font_size}"

    # def test_detection_pose(detection3dpc):
    """Test detection pose and frame information."""
    assert detection3dpc.pose.x == pytest.approx(-3.327, abs=0.1)
    assert detection3dpc.pose.y == pytest.approx(-0.202, abs=0.1)
    assert detection3dpc.pose.z == pytest.approx(0.160, abs=0.1)
    assert detection3dpc.pose.frame_id == "world", (
        f"Expected frame_id 'world', got '{detection3dpc.pose.frame_id}'"
    )



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Demo script to test agent message publishing and agentspy reception."""

import time

from langchain_core.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)

from dimos.protocol.pubsub import lcm  # type: ignore[attr-defined]
from dimos.protocol.pubsub.lcmpubsub import PickleLCM


def test_publish_messages() -> None:
    """Publish test messages to verify agentspy is working."""
    print("Starting agent message publisher demo...")

    # Create transport
    transport = PickleLCM()
    topic = lcm.Topic("/agent")

    print(f"Publishing to topic: {topic}")

    # Test messages
    messages = [
        SystemMessage("System initialized for testing"),
        HumanMessage("Hello agent, can you help me?"),
        AIMessage(
            "Of course! I'm here to help.",
            tool_calls=[{"name": "get_info", "args": {"query": "test"}, "id": "1"}],
        ),
        ToolMessage(name="get_info", content="Test result: success", tool_call_id="1"),
        AIMessage("The test was successful!", metadata={"state": True}),
    ]

    # Publish messages with delays
    for i, msg in enumerate(messages):
        print(f"\nPublishing message {i + 1}: {type(msg).__name__}")
        print(f"Content: {msg.content if hasattr(msg, 'content') else msg}")

        transport.publish(topic, msg)
        time.sleep(1)  # Wait 1 second between messages

    print("\nAll messages published! Check agentspy to see if they were received.")
    print("Keeping publisher alive for 10 more seconds...")
    time.sleep(10)


if __name__ == "__main__":
    test_publish_messages()



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Minimal TensorZero test to get it working."""

from pathlib import Path

from dotenv import load_dotenv
from openai import OpenAI
from tensorzero import patch_openai_client

load_dotenv()

# Create minimal config
config_dir = Path("/tmp/tz_test")
config_dir.mkdir(exist_ok=True)
config_path = config_dir / "tensorzero.toml"

# Minimal config based on TensorZero docs
config = """
[models.gpt_4o_mini]
routing = ["openai"]

[models.gpt_4o_mini.providers.openai]
type = "openai"
model_name = "gpt-4o-mini"

[functions.my_function]
type = "chat"

[functions.my_function.variants.my_variant]
type = "chat_completion"
model = "gpt_4o_mini"
"""

with open(config_path, "w") as f:
    f.write(config)

print(f"Created config at {config_path}")

# Create OpenAI client
client = OpenAI()

# Patch with TensorZero
try:
    patch_openai_client(
        client,
        clickhouse_url=None,  # In-memory
        config_file=str(config_path),
        async_setup=False,
    )
    print("✅ TensorZero initialized successfully!")
except Exception as e:
    print(f"❌ Failed to initialize TensorZero: {e}")
    exit(1)

# Test basic inference
print("\nTesting basic inference...")
try:
    response = client.chat.completions.create(
        model="tensorzero::function_name::my_function",
        messages=[{"role": "user", "content": "What is 2+2?"}],
        temperature=0.0,
        max_tokens=10,
    )

    content = response.choices[0].message.content
    print(f"Response: {content}")
    print("✅ Basic inference worked!")

except Exception as e:
    print(f"❌ Basic inference failed: {e}")
    import traceback

    traceback.print_exc()

print("\nTesting streaming...")
try:
    stream = client.chat.completions.create(
        model="tensorzero::function_name::my_function",
        messages=[{"role": "user", "content": "Count from 1 to 3"}],
        temperature=0.0,
        max_tokens=20,
        stream=True,
    )

    print("Stream response: ", end="", flush=True)
    for chunk in stream:
        if chunk.choices[0].delta.content:
            print(chunk.choices[0].delta.content, end="", flush=True)
    print("\n✅ Streaming worked!")

except Exception as e:
    print(f"\n❌ Streaming failed: {e}")



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from collections.abc import Generator
import datetime
import time

import pytest  # type: ignore[import-not-found]

from dimos.core import Module, rpc
from dimos.msgs.sensor_msgs import Image
from dimos.protocol.skill.coordinator import SkillCoordinator
from dimos.protocol.skill.skill import skill
from dimos.protocol.skill.type import Output, Reducer, Stream
from dimos.utils.data import get_data


class SkillContainerTest(Module):
    @rpc
    def start(self) -> None:
        super().start()

    @rpc
    def stop(self) -> None:
        super().stop()

    @skill()
    def add(self, x: int, y: int) -> int:
        """adds x and y."""
        time.sleep(2)
        return x + y

    @skill()
    def delayadd(self, x: int, y: int) -> int:
        """waits 0.3 seconds before adding x and y."""
        time.sleep(0.3)
        return x + y

    @skill(stream=Stream.call_agent, reducer=Reducer.all)  # type: ignore[arg-type]
    def counter(self, count_to: int, delay: float | None = 0.05) -> Generator[int, None, None]:
        """Counts from 1 to count_to, with an optional delay between counts."""
        for i in range(1, count_to + 1):
            if delay is not None and delay > 0:
                time.sleep(delay)
            yield i

    @skill(stream=Stream.passive, reducer=Reducer.sum)  # type: ignore[arg-type]
    def counter_passive_sum(
        self, count_to: int, delay: float | None = 0.05
    ) -> Generator[int, None, None]:
        """Counts from 1 to count_to, with an optional delay between counts."""
        for i in range(1, count_to + 1):
            if delay is not None and delay > 0:
                time.sleep(delay)
            yield i

    @skill(stream=Stream.passive, reducer=Reducer.latest)  # type: ignore[arg-type]
    def current_time(self, frequency: float | None = 10) -> Generator[str, None, None]:
        """Provides current time."""
        while True:
            yield str(datetime.datetime.now())
            if frequency is not None:
                time.sleep(1 / frequency)

    @skill(stream=Stream.passive, reducer=Reducer.latest)  # type: ignore[arg-type]
    def uptime_seconds(self, frequency: float | None = 10) -> Generator[float, None, None]:
        """Provides current uptime."""
        start_time = datetime.datetime.now()
        while True:
            yield (datetime.datetime.now() - start_time).total_seconds()
            if frequency is not None:
                time.sleep(1 / frequency)

    @skill()
    def current_date(self, frequency: float | None = 10) -> str:
        """Provides current date."""
        return str(datetime.datetime.now())

    @skill(output=Output.image)
    def take_photo(self) -> Image:
        """Takes a camera photo"""
        print("Taking photo...")
        img = Image.from_file(str(get_data("cafe-smol.jpg")))
        print("Photo taken.")
        return img


@pytest.mark.integration
@pytest.mark.asyncio  # type: ignore[untyped-decorator]
async def test_coordinator_parallel_calls() -> None:
    container = SkillContainerTest()
    skillCoordinator = SkillCoordinator()
    skillCoordinator.register_skills(container)

    skillCoordinator.start()
    skillCoordinator.call_skill("test-call-0", "add", {"args": [0, 2]})

    time.sleep(0.1)

    cnt = 0
    while await skillCoordinator.wait_for_updates(1):
        print(skillCoordinator)

        skillstates = skillCoordinator.generate_snapshot()

        skill_id = f"test-call-{cnt}"
        tool_msg = skillstates[skill_id].agent_encode()
        assert tool_msg.content == cnt + 2  # type: ignore[union-attr]

        cnt += 1
        if cnt < 5:
            skillCoordinator.call_skill(
                f"test-call-{cnt}-delay",
                "delayadd",
                {"args": [cnt, 2]},
            )
            skillCoordinator.call_skill(
                f"test-call-{cnt}",
                "add",
                {"args": [cnt, 2]},
            )

        await asyncio.sleep(0.1 * cnt)

    container.stop()
    skillCoordinator.stop()


@pytest.mark.integration
@pytest.mark.asyncio  # type: ignore[untyped-decorator]
async def test_coordinator_generator() -> None:
    container = SkillContainerTest()
    skillCoordinator = SkillCoordinator()
    skillCoordinator.register_skills(container)
    skillCoordinator.start()

    # here we call a skill that generates a sequence of messages
    skillCoordinator.call_skill("test-gen-0", "counter", {"args": [10]})
    skillCoordinator.call_skill("test-gen-1", "counter_passive_sum", {"args": [5]})
    skillCoordinator.call_skill("test-gen-2", "take_photo", {"args": []})

    # periodically agent is stopping it's thinking cycle and asks for updates
    while await skillCoordinator.wait_for_updates(2):
        print(skillCoordinator)
        agent_update = skillCoordinator.generate_snapshot(clear=True)
        print(agent_update)
        await asyncio.sleep(0.125)

    print("coordinator loop finished")
    print(skillCoordinator)
    container.stop()
    skillCoordinator.stop()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
TrajectoryStatus message type.

Status feedback for trajectory execution.
"""

from enum import IntEnum
from io import BytesIO
import struct
import time


class TrajectoryState(IntEnum):
    """States for trajectory execution."""

    IDLE = 0  # No trajectory, ready to accept
    EXECUTING = 1  # Currently executing trajectory
    COMPLETED = 2  # Trajectory finished successfully
    ABORTED = 3  # Trajectory was cancelled
    FAULT = 4  # Error occurred, requires reset()


class TrajectoryStatus:
    """
    Status of trajectory execution.

    Attributes:
        timestamp: When status was generated
        state: Current TrajectoryState
        progress: Progress 0.0 to 1.0
        time_elapsed: Seconds since trajectory start
        time_remaining: Estimated seconds remaining
        error: Error message if FAULT state (empty string otherwise)
    """

    msg_name = "trajectory_msgs.TrajectoryStatus"

    __slots__ = ["error", "progress", "state", "time_elapsed", "time_remaining", "timestamp"]

    def __init__(
        self,
        state: TrajectoryState = TrajectoryState.IDLE,
        progress: float = 0.0,
        time_elapsed: float = 0.0,
        time_remaining: float = 0.0,
        error: str = "",
        timestamp: float | None = None,
    ) -> None:
        """
        Initialize TrajectoryStatus.

        Args:
            state: Current execution state
            progress: Progress through trajectory (0.0 to 1.0)
            time_elapsed: Time since trajectory start (seconds)
            time_remaining: Estimated time remaining (seconds)
            error: Error message if in FAULT state
            timestamp: When status was generated (defaults to now)
        """
        self.timestamp = timestamp if timestamp is not None else time.time()
        self.state = state
        self.progress = progress
        self.time_elapsed = time_elapsed
        self.time_remaining = time_remaining
        self.error = error

    @property
    def state_name(self) -> str:
        """Get human-readable state name."""
        return self.state.name

    def is_done(self) -> bool:
        """Check if trajectory execution is finished (completed, aborted, or fault)."""
        return self.state in (
            TrajectoryState.COMPLETED,
            TrajectoryState.ABORTED,
            TrajectoryState.FAULT,
        )

    def is_active(self) -> bool:
        """Check if trajectory is currently executing."""
        return self.state == TrajectoryState.EXECUTING

    def lcm_encode(self) -> bytes:
        """Encode for LCM transport."""
        return self.encode()

    def encode(self) -> bytes:
        buf = BytesIO()
        buf.write(TrajectoryStatus._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf: BytesIO) -> None:
        # timestamp (double)
        buf.write(struct.pack(">d", self.timestamp))
        # state (int32)
        buf.write(struct.pack(">i", int(self.state)))
        # progress (double)
        buf.write(struct.pack(">d", self.progress))
        # time_elapsed (double)
        buf.write(struct.pack(">d", self.time_elapsed))
        # time_remaining (double)
        buf.write(struct.pack(">d", self.time_remaining))
        # error (string)
        error_bytes = self.error.encode("utf-8")
        buf.write(struct.pack(">i", len(error_bytes)))
        buf.write(error_bytes)

    @classmethod
    def lcm_decode(cls, data: bytes) -> "TrajectoryStatus":
        """Decode from LCM transport."""
        return cls.decode(data)

    @classmethod
    def decode(cls, data: bytes) -> "TrajectoryStatus":
        buf = BytesIO(data) if not hasattr(data, "read") else data
        if buf.read(8) != cls._get_packed_fingerprint():
            raise ValueError("Decode error: fingerprint mismatch")
        return cls._decode_one(buf)  # type: ignore[arg-type]

    @classmethod
    def _decode_one(cls, buf: BytesIO) -> "TrajectoryStatus":
        self = cls.__new__(cls)
        self.timestamp = struct.unpack(">d", buf.read(8))[0]
        self.state = TrajectoryState(struct.unpack(">i", buf.read(4))[0])
        self.progress = struct.unpack(">d", buf.read(8))[0]
        self.time_elapsed = struct.unpack(">d", buf.read(8))[0]
        self.time_remaining = struct.unpack(">d", buf.read(8))[0]
        error_len = struct.unpack(">i", buf.read(4))[0]
        self.error = buf.read(error_len).decode("utf-8")
        return self

    _packed_fingerprint = None

    @classmethod
    def _get_hash_recursive(cls, parents):  # type: ignore[no-untyped-def]
        if cls in parents:
            return 0
        return 0x3C4D5E6F708192A3 & 0xFFFFFFFFFFFFFFFF

    @classmethod
    def _get_packed_fingerprint(cls) -> bytes:
        if cls._packed_fingerprint is None:
            cls._packed_fingerprint = struct.pack(">Q", cls._get_hash_recursive([]))  # type: ignore[no-untyped-call]
        return cls._packed_fingerprint

    def __str__(self) -> str:
        return f"TrajectoryStatus({self.state_name}, progress={self.progress:.1%})"

    def __repr__(self) -> str:
        return (
            f"TrajectoryStatus(state={self.state_name}, progress={self.progress}, "
            f"time_elapsed={self.time_elapsed}, time_remaining={self.time_remaining}, "
            f"error='{self.error}')"
        )



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CLI entry point for Dimensional MCP Bridge.

Connects Claude Code (or other MCP clients) to a running DimOS agent.

Usage:
    python -m dimos.protocol.mcp  # Bridge to running DimOS on default port
"""

from __future__ import annotations

import asyncio

from dimos.protocol.mcp.bridge import main as bridge_main


def main() -> None:
    """Main entry point - connects to running DimOS via bridge."""
    asyncio.run(bridge_main())


if __name__ == "__main__":
    main()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import numpy as np
import pytest

try:
    from builtin_interfaces.msg import Time as ROSTime
    from geometry_msgs.msg import (
        Point as ROSPoint,
        Pose as ROSPose,
        PoseWithCovariance as ROSPoseWithCovariance,
        Quaternion as ROSQuaternion,
        Twist as ROSTwist,
        TwistWithCovariance as ROSTwistWithCovariance,
        Vector3 as ROSVector3,
    )
    from nav_msgs.msg import Odometry as ROSOdometry
    from std_msgs.msg import Header as ROSHeader
except ImportError:
    ROSTwist = None
    ROSHeader = None
    ROSPose = None
    ROSPoseWithCovariance = None
    ROSQuaternion = None
    ROSOdometry = None
    ROSPoint = None
    ROSTime = None
    ROSTwistWithCovariance = None
    ROSVector3 = None


from dimos.msgs.geometry_msgs.Pose import Pose
from dimos.msgs.geometry_msgs.PoseWithCovariance import PoseWithCovariance
from dimos.msgs.geometry_msgs.Twist import Twist
from dimos.msgs.geometry_msgs.TwistWithCovariance import TwistWithCovariance
from dimos.msgs.geometry_msgs.Vector3 import Vector3
from dimos.msgs.nav_msgs.Odometry import Odometry


def test_odometry_default_init() -> None:
    """Test default initialization."""
    if ROSVector3 is None:
        pytest.skip("ROS not available")
    if ROSTwistWithCovariance is None:
        pytest.skip("ROS not available")
    if ROSTime is None:
        pytest.skip("ROS not available")
    if ROSPoint is None:
        pytest.skip("ROS not available")
    if ROSOdometry is None:
        pytest.skip("ROS not available")
    if ROSQuaternion is None:
        pytest.skip("ROS not available")
    if ROSPoseWithCovariance is None:
        pytest.skip("ROS not available")
    if ROSPose is None:
        pytest.skip("ROS not available")
    if ROSHeader is None:
        pytest.skip("ROS not available")
    if ROSTwist is None:
        pytest.skip("ROS not available")
    odom = Odometry()

    # Should have current timestamp
    assert odom.ts > 0
    assert odom.frame_id == ""
    assert odom.child_frame_id == ""

    # Pose should be at origin with identity orientation
    assert odom.pose.position.x == 0.0
    assert odom.pose.position.y == 0.0
    assert odom.pose.position.z == 0.0
    assert odom.pose.orientation.w == 1.0

    # Twist should be zero
    assert odom.twist.linear.x == 0.0
    assert odom.twist.linear.y == 0.0
    assert odom.twist.linear.z == 0.0
    assert odom.twist.angular.x == 0.0
    assert odom.twist.angular.y == 0.0
    assert odom.twist.angular.z == 0.0

    # Covariances should be zero
    assert np.all(odom.pose.covariance == 0.0)
    assert np.all(odom.twist.covariance == 0.0)


def test_odometry_with_frames() -> None:
    """Test initialization with frame IDs."""
    ts = 1234567890.123456
    frame_id = "odom"
    child_frame_id = "base_link"

    odom = Odometry(ts=ts, frame_id=frame_id, child_frame_id=child_frame_id)

    assert odom.ts == ts
    assert odom.frame_id == frame_id
    assert odom.child_frame_id == child_frame_id


def test_odometry_with_pose_and_twist() -> None:
    """Test initialization with pose and twist."""
    pose = Pose(1.0, 2.0, 3.0, 0.1, 0.2, 0.3, 0.9)
    twist = Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1))

    odom = Odometry(ts=1000.0, frame_id="odom", child_frame_id="base_link", pose=pose, twist=twist)

    assert odom.pose.pose.position.x == 1.0
    assert odom.pose.pose.position.y == 2.0
    assert odom.pose.pose.position.z == 3.0
    assert odom.twist.twist.linear.x == 0.5
    assert odom.twist.twist.angular.z == 0.1


def test_odometry_with_covariances() -> None:
    """Test initialization with pose and twist with covariances."""
    pose = Pose(1.0, 2.0, 3.0)
    pose_cov = np.arange(36, dtype=float)
    pose_with_cov = PoseWithCovariance(pose, pose_cov)

    twist = Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1))
    twist_cov = np.arange(36, 72, dtype=float)
    twist_with_cov = TwistWithCovariance(twist, twist_cov)

    odom = Odometry(
        ts=1000.0,
        frame_id="odom",
        child_frame_id="base_link",
        pose=pose_with_cov,
        twist=twist_with_cov,
    )

    assert odom.pose.position.x == 1.0
    assert np.array_equal(odom.pose.covariance, pose_cov)
    assert odom.twist.linear.x == 0.5
    assert np.array_equal(odom.twist.covariance, twist_cov)


def test_odometry_copy_constructor() -> None:
    """Test copy constructor."""
    original = Odometry(
        ts=1000.0,
        frame_id="odom",
        child_frame_id="base_link",
        pose=Pose(1.0, 2.0, 3.0),
        twist=Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1)),
    )

    copy = Odometry(original)

    assert copy == original
    assert copy is not original
    assert copy.pose is not original.pose
    assert copy.twist is not original.twist


def test_odometry_dict_init() -> None:
    """Test initialization from dictionary."""
    odom_dict = {
        "ts": 1000.0,
        "frame_id": "odom",
        "child_frame_id": "base_link",
        "pose": Pose(1.0, 2.0, 3.0),
        "twist": Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1)),
    }

    odom = Odometry(odom_dict)

    assert odom.ts == 1000.0
    assert odom.frame_id == "odom"
    assert odom.child_frame_id == "base_link"
    assert odom.pose.position.x == 1.0
    assert odom.twist.linear.x == 0.5


def test_odometry_properties() -> None:
    """Test convenience properties."""
    pose = Pose(1.0, 2.0, 3.0, 0.1, 0.2, 0.3, 0.9)
    twist = Twist(Vector3(0.5, 0.6, 0.7), Vector3(0.1, 0.2, 0.3))

    odom = Odometry(ts=1000.0, frame_id="odom", child_frame_id="base_link", pose=pose, twist=twist)

    # Position properties
    assert odom.x == 1.0
    assert odom.y == 2.0
    assert odom.z == 3.0
    assert odom.position.x == 1.0
    assert odom.position.y == 2.0
    assert odom.position.z == 3.0

    # Orientation properties
    assert odom.orientation.x == 0.1
    assert odom.orientation.y == 0.2
    assert odom.orientation.z == 0.3
    assert odom.orientation.w == 0.9

    # Velocity properties
    assert odom.vx == 0.5
    assert odom.vy == 0.6
    assert odom.vz == 0.7
    assert odom.linear_velocity.x == 0.5
    assert odom.linear_velocity.y == 0.6
    assert odom.linear_velocity.z == 0.7

    # Angular velocity properties
    assert odom.wx == 0.1
    assert odom.wy == 0.2
    assert odom.wz == 0.3
    assert odom.angular_velocity.x == 0.1
    assert odom.angular_velocity.y == 0.2
    assert odom.angular_velocity.z == 0.3

    # Euler angles
    assert odom.roll == pose.roll
    assert odom.pitch == pose.pitch
    assert odom.yaw == pose.yaw


def test_odometry_str_repr() -> None:
    """Test string representations."""
    odom = Odometry(
        ts=1234567890.123456,
        frame_id="odom",
        child_frame_id="base_link",
        pose=Pose(1.234, 2.567, 3.891),
        twist=Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1)),
    )

    repr_str = repr(odom)
    assert "Odometry" in repr_str
    assert "1234567890.123456" in repr_str
    assert "odom" in repr_str
    assert "base_link" in repr_str

    str_repr = str(odom)
    assert "Odometry" in str_repr
    assert "odom -> base_link" in str_repr
    assert "1.234" in str_repr
    assert "0.500" in str_repr


def test_odometry_equality() -> None:
    """Test equality comparison."""
    odom1 = Odometry(
        ts=1000.0,
        frame_id="odom",
        child_frame_id="base_link",
        pose=Pose(1.0, 2.0, 3.0),
        twist=Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1)),
    )

    odom2 = Odometry(
        ts=1000.0,
        frame_id="odom",
        child_frame_id="base_link",
        pose=Pose(1.0, 2.0, 3.0),
        twist=Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1)),
    )

    odom3 = Odometry(
        ts=1000.0,
        frame_id="odom",
        child_frame_id="base_link",
        pose=Pose(1.1, 2.0, 3.0),  # Different position
        twist=Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1)),
    )

    assert odom1 == odom2
    assert odom1 != odom3
    assert odom1 != "not an odometry"


def test_odometry_lcm_encode_decode() -> None:
    """Test LCM encoding and decoding."""
    pose = Pose(1.0, 2.0, 3.0, 0.1, 0.2, 0.3, 0.9)
    pose_cov = np.arange(36, dtype=float)
    twist = Twist(Vector3(0.5, 0.6, 0.7), Vector3(0.1, 0.2, 0.3))
    twist_cov = np.arange(36, 72, dtype=float)

    source = Odometry(
        ts=1234567890.123456,
        frame_id="odom",
        child_frame_id="base_link",
        pose=PoseWithCovariance(pose, pose_cov),
        twist=TwistWithCovariance(twist, twist_cov),
    )

    # Encode and decode
    binary_msg = source.lcm_encode()
    decoded = Odometry.lcm_decode(binary_msg)

    # Check values (allowing for timestamp precision loss)
    assert abs(decoded.ts - source.ts) < 1e-6
    assert decoded.frame_id == source.frame_id
    assert decoded.child_frame_id == source.child_frame_id
    assert decoded.pose == source.pose
    assert decoded.twist == source.twist


@pytest.mark.ros
def test_odometry_from_ros_msg() -> None:
    """Test creating from ROS message."""
    ros_msg = ROSOdometry()

    # Set header
    ros_msg.header = ROSHeader()
    ros_msg.header.stamp = ROSTime()
    ros_msg.header.stamp.sec = 1234567890
    ros_msg.header.stamp.nanosec = 123456000
    ros_msg.header.frame_id = "odom"
    ros_msg.child_frame_id = "base_link"

    # Set pose with covariance
    ros_msg.pose = ROSPoseWithCovariance()
    ros_msg.pose.pose = ROSPose()
    ros_msg.pose.pose.position = ROSPoint(x=1.0, y=2.0, z=3.0)
    ros_msg.pose.pose.orientation = ROSQuaternion(x=0.1, y=0.2, z=0.3, w=0.9)
    ros_msg.pose.covariance = [float(i) for i in range(36)]

    # Set twist with covariance
    ros_msg.twist = ROSTwistWithCovariance()
    ros_msg.twist.twist = ROSTwist()
    ros_msg.twist.twist.linear = ROSVector3(x=0.5, y=0.6, z=0.7)
    ros_msg.twist.twist.angular = ROSVector3(x=0.1, y=0.2, z=0.3)
    ros_msg.twist.covariance = [float(i) for i in range(36, 72)]

    odom = Odometry.from_ros_msg(ros_msg)

    assert odom.ts == 1234567890.123456
    assert odom.frame_id == "odom"
    assert odom.child_frame_id == "base_link"
    assert odom.pose.position.x == 1.0
    assert odom.twist.linear.x == 0.5
    assert np.array_equal(odom.pose.covariance, np.arange(36))
    assert np.array_equal(odom.twist.covariance, np.arange(36, 72))


@pytest.mark.ros
def test_odometry_to_ros_msg() -> None:
    """Test converting to ROS message."""
    pose = Pose(1.0, 2.0, 3.0, 0.1, 0.2, 0.3, 0.9)
    pose_cov = np.arange(36, dtype=float)
    twist = Twist(Vector3(0.5, 0.6, 0.7), Vector3(0.1, 0.2, 0.3))
    twist_cov = np.arange(36, 72, dtype=float)

    odom = Odometry(
        ts=1234567890.567890,
        frame_id="odom",
        child_frame_id="base_link",
        pose=PoseWithCovariance(pose, pose_cov),
        twist=TwistWithCovariance(twist, twist_cov),
    )

    ros_msg = odom.to_ros_msg()

    assert isinstance(ros_msg, ROSOdometry)
    assert ros_msg.header.frame_id == "odom"
    assert ros_msg.header.stamp.sec == 1234567890
    assert abs(ros_msg.header.stamp.nanosec - 567890000) < 100  # Allow small rounding error
    assert ros_msg.child_frame_id == "base_link"

    # Check pose
    assert ros_msg.pose.pose.position.x == 1.0
    assert ros_msg.pose.pose.position.y == 2.0
    assert ros_msg.pose.pose.position.z == 3.0
    assert ros_msg.pose.pose.orientation.x == 0.1
    assert ros_msg.pose.pose.orientation.y == 0.2
    assert ros_msg.pose.pose.orientation.z == 0.3
    assert ros_msg.pose.pose.orientation.w == 0.9
    assert list(ros_msg.pose.covariance) == list(range(36))

    # Check twist
    assert ros_msg.twist.twist.linear.x == 0.5
    assert ros_msg.twist.twist.linear.y == 0.6
    assert ros_msg.twist.twist.linear.z == 0.7
    assert ros_msg.twist.twist.angular.x == 0.1
    assert ros_msg.twist.twist.angular.y == 0.2
    assert ros_msg.twist.twist.angular.z == 0.3
    assert list(ros_msg.twist.covariance) == list(range(36, 72))


@pytest.mark.ros
def test_odometry_ros_roundtrip() -> None:
    """Test round-trip conversion with ROS messages."""
    pose = Pose(1.5, 2.5, 3.5, 0.15, 0.25, 0.35, 0.85)
    pose_cov = np.random.rand(36)
    twist = Twist(Vector3(0.55, 0.65, 0.75), Vector3(0.15, 0.25, 0.35))
    twist_cov = np.random.rand(36)

    original = Odometry(
        ts=2147483647.987654,  # Max int32 value for ROS Time.sec
        frame_id="world",
        child_frame_id="robot",
        pose=PoseWithCovariance(pose, pose_cov),
        twist=TwistWithCovariance(twist, twist_cov),
    )

    ros_msg = original.to_ros_msg()
    restored = Odometry.from_ros_msg(ros_msg)

    # Check values (allowing for timestamp precision loss)
    assert abs(restored.ts - original.ts) < 1e-6
    assert restored.frame_id == original.frame_id
    assert restored.child_frame_id == original.child_frame_id
    assert restored.pose == original.pose
    assert restored.twist == original.twist


def test_odometry_zero_timestamp() -> None:
    """Test that zero timestamp gets replaced with current time."""
    odom = Odometry(ts=0.0)

    # Should have been replaced with current time
    assert odom.ts > 0
    assert odom.ts <= time.time()


def test_odometry_with_just_pose() -> None:
    """Test initialization with just a Pose (no covariance)."""
    pose = Pose(1.0, 2.0, 3.0)

    odom = Odometry(pose=pose)

    assert odom.pose.position.x == 1.0
    assert odom.pose.position.y == 2.0
    assert odom.pose.position.z == 3.0
    assert np.all(odom.pose.covariance == 0.0)  # Should have zero covariance
    assert np.all(odom.twist.covariance == 0.0)  # Twist should also be zero


def test_odometry_with_just_twist() -> None:
    """Test initialization with just a Twist (no covariance)."""
    twist = Twist(Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.1))

    odom = Odometry(twist=twist)

    assert odom.twist.linear.x == 0.5
    assert odom.twist.angular.z == 0.1
    assert np.all(odom.twist.covariance == 0.0)  # Should have zero covariance
    assert np.all(odom.pose.covariance == 0.0)  # Pose should also be zero


@pytest.mark.ros
@pytest.mark.parametrize(
    "frame_id,child_frame_id",
    [
        ("odom", "base_link"),
        ("map", "odom"),
        ("world", "robot"),
        ("base_link", "camera_link"),
        ("", ""),  # Empty frames
    ],
)
def test_odometry_frame_combinations(frame_id, child_frame_id) -> None:
    """Test various frame ID combinations."""
    odom = Odometry(frame_id=frame_id, child_frame_id=child_frame_id)

    assert odom.frame_id == frame_id
    assert odom.child_frame_id == child_frame_id

    # Test roundtrip through ROS
    ros_msg = odom.to_ros_msg()
    assert ros_msg.header.frame_id == frame_id
    assert ros_msg.child_frame_id == child_frame_id

    restored = Odometry.from_ros_msg(ros_msg)
    assert restored.frame_id == frame_id
    assert restored.child_frame_id == child_frame_id


def test_odometry_typical_robot_scenario() -> None:
    """Test a typical robot odometry scenario."""
    # Robot moving forward at 0.5 m/s with slight rotation
    odom = Odometry(
        ts=1000.0,
        frame_id="odom",
        child_frame_id="base_footprint",
        pose=Pose(10.0, 5.0, 0.0, 0.0, 0.0, np.sin(0.1), np.cos(0.1)),  # 0.2 rad yaw
        twist=Twist(
            Vector3(0.5, 0.0, 0.0), Vector3(0.0, 0.0, 0.05)
        ),  # Moving forward, turning slightly
    )

    # Check we can access all the typical properties
    assert odom.x == 10.0
    assert odom.y == 5.0
    assert odom.z == 0.0
    assert abs(odom.yaw - 0.2) < 0.01  # Approximately 0.2 radians
    assert odom.vx == 0.5  # Forward velocity
    assert odom.wz == 0.05  # Yaw rate



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joint Trajectory Generator

Generates time-parameterized joint trajectories from waypoints using
trapezoidal velocity profiles.

Trapezoidal Profile:
    velocity
       ^
       |    ____________________
       |   /                    \
       |  /                      \
       | /                        \
       |/                          \
       +------------------------------> time
        accel    cruise      decel
"""

import math

from dimos.msgs.trajectory_msgs import JointTrajectory, TrajectoryPoint


class JointTrajectoryGenerator:
    """
    Generates joint trajectories with trapezoidal velocity profiles.

    For each segment between waypoints:
    1. Determines the limiting joint (one that takes longest)
    2. Applies trapezoidal velocity profile based on limits
    3. Scales other joints to complete in the same time
    4. Generates trajectory points with proper timing

    Usage:
        generator = JointTrajectoryGenerator(num_joints=6)
        generator.set_limits(max_velocity=1.0, max_acceleration=2.0)
        trajectory = generator.generate(waypoints)
    """

    def __init__(
        self,
        num_joints: int = 6,
        max_velocity: list[float] | float = 1.0,
        max_acceleration: list[float] | float = 2.0,
        points_per_segment: int = 50,
    ) -> None:
        """
        Initialize trajectory generator.

        Args:
            num_joints: Number of joints
            max_velocity: rad/s (single value applies to all joints, or per-joint list)
            max_acceleration: rad/s^2 (single value or per-joint list)
            points_per_segment: Number of intermediate points per waypoint segment
        """
        self.num_joints = num_joints
        self.points_per_segment = points_per_segment

        # Initialize limits
        self.max_velocity: list[float] = []
        self.max_acceleration: list[float] = []
        self.set_limits(max_velocity, max_acceleration)

    def set_limits(
        self,
        max_velocity: list[float] | float,
        max_acceleration: list[float] | float,
    ) -> None:
        """
        Set velocity and acceleration limits.

        Args:
            max_velocity: rad/s (single value applies to all joints, or per-joint)
            max_acceleration: rad/s^2 (single value or per-joint)
        """
        if isinstance(max_velocity, (int, float)):
            self.max_velocity = [float(max_velocity)] * self.num_joints
        else:
            self.max_velocity = list(max_velocity)

        if isinstance(max_acceleration, (int, float)):
            self.max_acceleration = [float(max_acceleration)] * self.num_joints
        else:
            self.max_acceleration = list(max_acceleration)

    def generate(self, waypoints: list[list[float]]) -> JointTrajectory:
        """
        Generate a trajectory through waypoints with trapezoidal velocity profile.

        Args:
            waypoints: List of joint positions [q1, q2, ..., qn] in radians
                       First waypoint is start, last is goal

        Returns:
            JointTrajectory with time-parameterized points
        """
        if not waypoints or len(waypoints) < 2:
            raise ValueError("Need at least 2 waypoints")

        all_points: list[TrajectoryPoint] = []
        current_time = 0.0

        # Add first waypoint
        all_points.append(
            TrajectoryPoint(
                time_from_start=0.0,
                positions=list(waypoints[0]),
                velocities=[0.0] * self.num_joints,
            )
        )

        # Process each segment
        for i in range(len(waypoints) - 1):
            start = waypoints[i]
            end = waypoints[i + 1]

            # Generate segment with trapezoidal profile
            segment_points, segment_duration = self._generate_segment(start, end, current_time)

            # Add points (skip first as it duplicates previous endpoint)
            all_points.extend(segment_points[1:])
            current_time += segment_duration

        return JointTrajectory(points=all_points)

    def _generate_segment(
        self,
        start: list[float],
        end: list[float],
        start_time: float,
    ) -> tuple[list[TrajectoryPoint], float]:
        """
        Generate trajectory points for a single segment using trapezoidal profile.

        Args:
            start: Starting joint positions
            end: Ending joint positions
            start_time: Time offset for this segment

        Returns:
            Tuple of (list of TrajectoryPoints, segment duration)
        """
        # Calculate displacement for each joint
        displacements = [end[j] - start[j] for j in range(self.num_joints)]

        # Find the limiting joint (one that takes longest)
        segment_duration = 0.0
        for j in range(self.num_joints):
            t = self._compute_trapezoidal_time(
                abs(displacements[j]),
                self.max_velocity[j],
                self.max_acceleration[j],
            )
            segment_duration = max(segment_duration, t)

        # Ensure minimum duration
        segment_duration = max(segment_duration, 0.01)

        # Generate points along the segment
        points: list[TrajectoryPoint] = []

        for i in range(self.points_per_segment + 1):
            # Normalized time [0, 1]
            s = i / self.points_per_segment
            t = start_time + s * segment_duration

            # Compute position and velocity for each joint
            positions = []
            velocities = []

            for j in range(self.num_joints):
                # Compute scaled limits for this joint to fit in segment_duration
                v_scaled, a_scaled = self._compute_scaled_limits(
                    abs(displacements[j]),
                    segment_duration,
                    self.max_velocity[j],
                    self.max_acceleration[j],
                )

                pos, vel = self._trapezoidal_interpolate(
                    s,
                    start[j],
                    end[j],
                    segment_duration,
                    v_scaled,
                    a_scaled,
                )
                positions.append(pos)
                velocities.append(vel)

            points.append(
                TrajectoryPoint(
                    time_from_start=t,
                    positions=positions,
                    velocities=velocities,
                )
            )

        return points, segment_duration

    def _compute_trapezoidal_time(
        self,
        distance: float,
        v_max: float,
        a_max: float,
    ) -> float:
        """
        Compute time to travel a distance with trapezoidal velocity profile.

        Two cases:
        1. Triangle profile: Can't reach v_max (short distance)
        2. Trapezoidal profile: Reaches v_max with cruise phase

        Args:
            distance: Absolute distance to travel
            v_max: Maximum velocity
            a_max: Maximum acceleration

        Returns:
            Time to complete the motion
        """
        if distance < 1e-9:
            return 0.0

        # Time to accelerate to v_max
        t_accel = v_max / a_max

        # Distance covered during accel + decel (both at a_max)
        d_accel = 0.5 * a_max * t_accel**2
        d_total_ramp = 2 * d_accel  # accel + decel

        if distance <= d_total_ramp:
            # Triangle profile - can't reach v_max
            # d = 2 * (0.5 * a * t^2) = a * t^2
            # t = sqrt(d / a)
            t_ramp = math.sqrt(distance / a_max)
            return 2 * t_ramp
        else:
            # Trapezoidal profile - has cruise phase
            d_cruise = distance - d_total_ramp
            t_cruise = d_cruise / v_max
            return 2 * t_accel + t_cruise

    def _compute_scaled_limits(
        self,
        distance: float,
        duration: float,
        v_max: float,
        a_max: float,
    ) -> tuple[float, float]:
        """
        Compute scaled velocity and acceleration to travel distance in given duration.

        This scales down the profile so the joint travels its distance in the
        same time as the limiting joint.

        Args:
            distance: Absolute distance to travel
            duration: Required duration (from limiting joint)
            v_max: Maximum velocity limit
            a_max: Maximum acceleration limit

        Returns:
            Tuple of (scaled_velocity, scaled_acceleration)
        """
        if distance < 1e-9 or duration < 1e-9:
            return v_max, a_max

        # Compute optimal time for this joint
        t_opt = self._compute_trapezoidal_time(distance, v_max, a_max)

        if t_opt >= duration - 1e-9:
            # This is the limiting joint or close to it
            return v_max, a_max

        # Need to scale down to fit in longer duration
        # Use simple scaling: scale both v and a by the same factor
        # This preserves the profile shape
        scale = t_opt / duration

        # For a symmetric trapezoidal/triangular profile:
        # If we scale time by k, we need to scale velocity by 1/k
        # But we also need to ensure we travel the same distance

        # Simpler approach: compute the average velocity needed
        distance / duration

        # For trapezoidal profile, v_avg = v_peak * (1 - t_accel/duration)
        # For simplicity, use a heuristic: scale velocity so trajectory fits

        # Check if we can use a triangle profile
        # Triangle: d = 0.5 * v_peak * T, so v_peak = 2 * d / T
        v_peak_triangle = 2 * distance / duration
        a_for_triangle = 4 * distance / (duration * duration)

        if v_peak_triangle <= v_max and a_for_triangle <= a_max:
            # Use triangle profile with these params
            return v_peak_triangle, a_for_triangle

        # Use trapezoidal with reduced velocity
        # Solve: distance = v * t_cruise + v^2/a
        # where t_cruise = duration - 2*v/a
        # This is complex, so use iterative scaling
        v_scaled = v_max * scale
        a_scaled = a_max * scale * scale  # acceleration scales with square of time scale

        # Verify and adjust
        t_check = self._compute_trapezoidal_time(distance, v_scaled, a_scaled)
        if abs(t_check - duration) > 0.01 * duration:
            # Fallback: use triangle profile scaled to fit
            v_scaled = 2 * distance / duration
            a_scaled = 4 * distance / (duration * duration)

        return min(v_scaled, v_max), min(a_scaled, a_max)

    def _trapezoidal_interpolate(
        self,
        s: float,
        start: float,
        end: float,
        duration: float,
        v_max: float,
        a_max: float,
    ) -> tuple[float, float]:
        """
        Interpolate position and velocity using trapezoidal profile.

        Args:
            s: Normalized time [0, 1]
            start: Start position
            end: End position
            duration: Total segment duration
            v_max: Max velocity for this joint (scaled)
            a_max: Max acceleration for this joint (scaled)

        Returns:
            Tuple of (position, velocity)
        """
        distance = abs(end - start)
        direction = 1.0 if end >= start else -1.0

        if distance < 1e-9 or duration < 1e-9:
            return end, 0.0

        # Handle endpoint exactly
        if s >= 1.0 - 1e-9:
            return end, 0.0
        if s <= 1e-9:
            return start, 0.0

        # Current time
        t = s * duration

        # Compute profile parameters for this joint
        t_accel = v_max / a_max if a_max > 1e-9 else duration / 2
        d_accel = 0.5 * a_max * t_accel**2
        d_total_ramp = 2 * d_accel

        if distance <= d_total_ramp + 1e-9:
            # Triangle profile
            t_peak = duration / 2
            v_peak = 2 * distance / duration
            a_eff = v_peak / t_peak if t_peak > 1e-9 else a_max

            if t <= t_peak:
                # Accelerating
                pos_offset = 0.5 * a_eff * t * t
                vel = direction * a_eff * t
            else:
                # Decelerating
                dt = t - t_peak
                pos_offset = distance / 2 + v_peak * dt - 0.5 * a_eff * dt * dt
                vel = direction * max(0.0, v_peak - a_eff * dt)
        else:
            # Trapezoidal profile
            d_cruise = distance - d_total_ramp
            t_cruise = d_cruise / v_max if v_max > 1e-9 else 0

            if t <= t_accel:
                # Accelerating phase
                pos_offset = 0.5 * a_max * t * t
                vel = direction * a_max * t
            elif t <= t_accel + t_cruise:
                # Cruise phase
                dt = t - t_accel
                pos_offset = d_accel + v_max * dt
                vel = direction * v_max
            else:
                # Decelerating phase
                dt = t - t_accel - t_cruise
                pos_offset = d_accel + d_cruise + v_max * dt - 0.5 * a_max * dt * dt
                vel = direction * max(0.0, v_max - a_max * dt)

        position = start + direction * pos_offset

        # Clamp to ensure we don't overshoot
        if direction > 0:
            position = min(position, end)
        else:
            position = max(position, end)

        return position, vel

    def preview(self, trajectory: JointTrajectory) -> str:
        """
        Generate a text preview of the trajectory.

        Args:
            trajectory: Generated trajectory to preview

        Returns:
            Formatted string showing trajectory details
        """
        lines = [
            "Trajectory Preview",
            "=" * 60,
            f"Duration: {trajectory.duration:.3f}s",
            f"Points: {len(trajectory.points)}",
            "",
            "Waypoints (time -> positions):",
            "-" * 60,
        ]

        # Show key points (first, last, and evenly spaced)
        indices = [0]
        step = max(1, len(trajectory.points) // 5)
        indices.extend(range(step, len(trajectory.points) - 1, step))
        indices.append(len(trajectory.points) - 1)
        indices = sorted(set(indices))

        for i in indices:
            pt = trajectory.points[i]
            pos_str = ", ".join(f"{p:+.3f}" for p in pt.positions)
            vel_str = ", ".join(f"{v:+.3f}" for v in pt.velocities)
            lines.append(f"  t={pt.time_from_start:6.3f}s: pos=[{pos_str}]")
            lines.append(f"           vel=[{vel_str}]")

        lines.append("-" * 60)
        return "\n".join(lines)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterator
import glob
import os
import pickle
from typing import cast, overload

from reactivex import from_iterable, interval, operators as ops
from reactivex.observable import Observable

from dimos.msgs.sensor_msgs import PointCloud2
from dimos.robot.unitree_webrtc.type.lidar import RawLidarMsg, pointcloud2_from_webrtc_lidar


class Mock:
    def __init__(self, root: str = "office", autocast: bool = True) -> None:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        self.root = os.path.join(current_dir, f"mockdata/{root}")
        self.autocast = autocast
        self.cnt = 0

    @overload
    def load(self, name: int | str, /) -> PointCloud2: ...
    @overload
    def load(self, *names: int | str) -> list[PointCloud2]: ...

    def load(self, *names: int | str) -> PointCloud2 | list[PointCloud2]:
        if len(names) == 1:
            return self.load_one(names[0])
        return list(map(lambda name: self.load_one(name), names))

    def load_one(self, name: int | str) -> PointCloud2:
        if isinstance(name, int):
            file_name = f"/lidar_data_{name:03d}.pickle"
        else:
            file_name = f"/{name}.pickle"

        full_path = self.root + file_name
        with open(full_path, "rb") as f:
            return pointcloud2_from_webrtc_lidar(cast("RawLidarMsg", pickle.load(f)))

    def iterate(self) -> Iterator[PointCloud2]:
        pattern = os.path.join(self.root, "lidar_data_*.pickle")
        print("loading data", pattern)
        for file_path in sorted(glob.glob(pattern)):
            basename = os.path.basename(file_path)
            filename = os.path.splitext(basename)[0]
            yield self.load_one(filename)

    def stream(self, rate_hz: float = 10.0):  # type: ignore[no-untyped-def]
        sleep_time = 1.0 / rate_hz

        return from_iterable(self.iterate()).pipe(
            ops.zip(interval(sleep_time)),
            ops.map(lambda x: x[0] if isinstance(x, tuple) else x),
        )

    def save_stream(self, observable: Observable[PointCloud2]):  # type: ignore[no-untyped-def]
        return observable.pipe(ops.map(lambda frame: self.save_one(frame)))  # type: ignore[no-untyped-call]

    def save(self, *frames):  # type: ignore[no-untyped-def]
        [self.save_one(frame) for frame in frames]  # type: ignore[no-untyped-call]
        return self.cnt

    def save_one(self, frame):  # type: ignore[no-untyped-def]
        file_name = f"/lidar_data_{self.cnt:03d}.pickle"
        full_path = self.root + file_name

        self.cnt += 1

        if os.path.isfile(full_path):
            raise Exception(f"file {full_path} exists")

        # Note: This saves the PointCloud2 directly. For raw message saving,
        # use the raw message before conversion.
        with open(full_path, "wb") as f:
            pickle.dump(frame, f)

        return self.cnt



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import pytest

from dimos.agents.agent import Agent
from dimos.agents.testing import MockModel
from dimos.protocol.skill.test_coordinator import SkillContainerTest


@pytest.fixture
def fixture_dir():
    return Path(__file__).parent / "fixtures"


@pytest.fixture
def potato_system_prompt() -> str:
    return "Your name is Mr. Potato, potatoes are bad at math. Use a tools if asked to calculate"


@pytest.fixture
def skill_container():
    container = SkillContainerTest()
    try:
        yield container
    finally:
        container.stop()


@pytest.fixture
def create_fake_agent(fixture_dir):
    agent = None

    def _agent_factory(*, system_prompt, skill_containers, fixture):
        mock_model = MockModel(json_path=fixture_dir / fixture)

        nonlocal agent
        agent = Agent(system_prompt=system_prompt, model_instance=mock_model)

        for skill_container in skill_containers:
            agent.register_skills(skill_container)

        agent.start()

        return agent

    try:
        yield _agent_factory
    finally:
        if agent:
            agent.stop()


@pytest.fixture
def create_potato_agent(potato_system_prompt, skill_container, fixture_dir):
    agent = None

    def _agent_factory(*, fixture):
        mock_model = MockModel(json_path=fixture_dir / fixture)

        nonlocal agent
        agent = Agent(system_prompt=potato_system_prompt, model_instance=mock_model)
        agent.register_skills(skill_container)
        agent.start()

        return agent

    try:
        yield _agent_factory
    finally:
        if agent:
            agent.stop()



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
import numpy as np
from open3d.geometry import PointCloud
import pytest

from dimos.core import LCMTransport
from dimos.mapping.occupancy.visualizations import visualize_occupancy_grid
from dimos.mapping.pointclouds.occupancy import (
    height_cost_occupancy,
    simple_occupancy,
)
from dimos.mapping.pointclouds.util import read_pointcloud
from dimos.msgs.nav_msgs.OccupancyGrid import OccupancyGrid
from dimos.msgs.sensor_msgs import PointCloud2
from dimos.msgs.sensor_msgs.Image import Image
from dimos.utils.data import get_data
from dimos.utils.testing.moment import OutputMoment
from dimos.utils.testing.test_moment import Go2Moment


@pytest.fixture
def apartment() -> PointCloud:
    return read_pointcloud(get_data("apartment") / "sum.ply")


@pytest.fixture
def big_office() -> PointCloud:
    return read_pointcloud(get_data("big_office.ply"))


@pytest.mark.parametrize(
    "occupancy_fn,output_name",
    [
        (simple_occupancy, "occupancy_simple.png"),
    ],
)
def test_occupancy(apartment: PointCloud, occupancy_fn, output_name: str) -> None:
    expected_image = cv2.imread(str(get_data(output_name)), cv2.IMREAD_GRAYSCALE)
    cloud = PointCloud2.from_numpy(np.asarray(apartment.points), frame_id="map")

    occupancy_grid = occupancy_fn(cloud)

    # Convert grid from -1..100 to 0..101 for PNG
    computed_image = (occupancy_grid.grid + 1).astype(np.uint8)

    np.testing.assert_array_equal(computed_image, expected_image)


@pytest.mark.parametrize(
    "occupancy_fn,output_name",
    [
        (height_cost_occupancy, "big_office_height_cost_occupancy.png"),
        (simple_occupancy, "big_office_simple_occupancy.png"),
    ],
)
def test_occupancy2(big_office, occupancy_fn, output_name):
    expected_image = Image.from_file(get_data(output_name))
    cloud = PointCloud2.from_numpy(np.asarray(big_office.points), frame_id="")

    occupancy_grid = occupancy_fn(cloud)

    actual = visualize_occupancy_grid(occupancy_grid, "rainbow")
    actual.ts = expected_image.ts
    np.testing.assert_array_equal(actual, expected_image)


class HeightCostMoment(Go2Moment):
    costmap: OutputMoment[OccupancyGrid] = OutputMoment(LCMTransport("/costmap", OccupancyGrid))


@pytest.fixture
def height_cost_moment():
    moment = HeightCostMoment()

    def get_moment(ts: float, publish: bool = True) -> HeightCostMoment:
        moment.seek(ts)
        if moment.lidar.value is not None:
            costmap = height_cost_occupancy(
                moment.lidar.value,
                resolution=0.05,
                can_pass_under=0.6,
                can_climb=0.15,
            )
            moment.costmap.set(costmap)
        if publish:
            moment.publish()
        return moment

    yield get_moment

    moment.stop()


def test_height_cost_occupancy_from_lidar(height_cost_moment) -> None:
    """Test height_cost_occupancy with real lidar data."""
    moment = height_cost_moment(1.0)

    costmap = moment.costmap.value
    assert costmap is not None

    # Basic sanity checks
    assert costmap.grid is not None
    assert costmap.width > 0
    assert costmap.height > 0

    # Costs should be in range -1 to 100 (-1 = unknown)
    assert costmap.grid.min() >= -1
    assert costmap.grid.max() <= 100

    # Check we have some unknown, some known
    known_mask = costmap.grid >= 0
    assert known_mask.sum() > 0, "Expected some known cells"
    assert (~known_mask).sum() > 0, "Expected some unknown cells"



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypedDict, cast

from dimos.constants import DEFAULT_CAPACITY_COLOR_IMAGE
from dimos.core import DimosCluster, LCMTransport, pSHMTransport
from dimos.hardware.sensors.camera import zed
from dimos.hardware.sensors.camera.module import CameraModule
from dimos.hardware.sensors.camera.webcam import Webcam
from dimos.msgs.geometry_msgs import (
    Quaternion,
    Transform,
    Vector3,
)
from dimos.msgs.sensor_msgs import CameraInfo
from dimos.navigation import rosnav
from dimos.navigation.rosnav import ROSNav
from dimos.robot import foxglove_bridge
from dimos.robot.unitree.connection import g1
from dimos.robot.unitree.connection.g1 import G1Connection
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


class G1ZedDeployResult(TypedDict):
    nav: ROSNav
    connection: G1Connection
    camera: CameraModule
    camerainfo: CameraInfo


def deploy_g1_monozed(dimos: DimosCluster) -> CameraModule:
    camera = cast(
        "CameraModule",
        dimos.deploy(  # type: ignore[attr-defined]
            CameraModule,
            frequency=4.0,
            transform=Transform(
                translation=Vector3(0.05, 0.0, 0.0),
                rotation=Quaternion.from_euler(Vector3(0.0, 0.0, 0.0)),
                frame_id="sensor",
                child_frame_id="camera_link",
            ),
            hardware=lambda: Webcam(
                camera_index=0,
                fps=5,
                stereo_slice="left",
                camera_info=zed.CameraInfo.SingleWebcam,
            ),
        ),
    )

    camera.color_image.transport = pSHMTransport(
        "/image", default_capacity=DEFAULT_CAPACITY_COLOR_IMAGE
    )
    camera.camera_info.transport = LCMTransport("/camera_info", CameraInfo)
    camera.start()
    return camera


def deploy(dimos: DimosCluster, ip: str):  # type: ignore[no-untyped-def]
    nav = rosnav.deploy(  # type: ignore[call-arg]
        dimos,
        sensor_to_base_link_transform=Transform(
            frame_id="sensor", child_frame_id="base_link", translation=Vector3(0.0, 0.0, 1.5)
        ),
    )
    connection = g1.deploy(dimos, ip, nav)
    zedcam = deploy_g1_monozed(dimos)

    foxglove_bridge.deploy(dimos)

    return {
        "nav": nav,
        "connection": connection,
        "camera": zedcam,
    }



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import pathlib
import tempfile
import time

from dotenv import load_dotenv
import pytest
from reactivex import operators as ops

from dimos import core
from dimos.core import Module, Out, rpc
from dimos.models.vl.openai import OpenAIVlModel
from dimos.msgs.sensor_msgs import Image
from dimos.perception.experimental.temporal_memory import TemporalMemory, TemporalMemoryConfig
from dimos.protocol import pubsub
from dimos.utils.data import get_data
from dimos.utils.logging_config import setup_logger
from dimos.utils.testing import TimedSensorReplay

# Load environment variables
load_dotenv()

logger = setup_logger()

pubsub.lcm.autoconf()


class VideoReplayModule(Module):
    """Module that replays video data from TimedSensorReplay."""

    video_out: Out[Image]

    def __init__(self, video_path: str) -> None:
        super().__init__()
        self.video_path = video_path

    @rpc
    def start(self) -> None:
        """Start replaying video data."""
        # Use TimedSensorReplay to replay video frames
        video_replay = TimedSensorReplay(self.video_path, autocast=Image.from_numpy)

        # Subscribe to the replay stream and publish to LCM
        self._disposables.add(
            video_replay.stream()
            .pipe(
                ops.sample(1),  # Sample every 1 second
                ops.take(10),  # Only take 10 frames total
            )
            .subscribe(self.video_out.publish)
        )

        logger.info("VideoReplayModule started")

    @rpc
    def stop(self) -> None:
        """Stop replaying video data."""
        # Stop all stream transports to clean up LCM loop threads
        for stream in list(self.outputs.values()):
            if stream.transport is not None and hasattr(stream.transport, "stop"):
                stream.transport.stop()
                stream._transport = None
        super().stop()
        logger.info("VideoReplayModule stopped")


@pytest.mark.skipif(bool(os.getenv("CI")), reason="LCM replay + dataset not CI-safe.")
@pytest.mark.skipif(not os.getenv("OPENAI_API_KEY"), reason="OPENAI_API_KEY not set.")
@pytest.mark.neverending
class TestTemporalMemoryModule:
    @pytest.fixture(scope="function")
    def temp_dir(self):
        """Create a temporary directory for test data."""
        temp_dir = tempfile.mkdtemp(prefix="temporal_memory_test_")
        yield temp_dir

    @pytest.fixture(scope="function")
    def dimos_cluster(self):
        """Create and cleanup Dimos cluster."""
        dimos = core.start(1)
        yield dimos
        dimos.close_all()

    @pytest.fixture(scope="function")
    def video_module(self, dimos_cluster):
        """Create and cleanup video replay module."""
        data_path = get_data("unitree_office_walk")
        video_path = os.path.join(data_path, "video")
        video_module = dimos_cluster.deploy(VideoReplayModule, video_path)
        video_module.video_out.transport = core.LCMTransport("/test_video", Image)
        yield video_module
        try:
            video_module.stop()
        except Exception as e:
            logger.warning(f"Failed to stop video_module: {e}")

    @pytest.fixture(scope="function")
    def temporal_memory(self, dimos_cluster, temp_dir):
        """Create and cleanup temporal memory module."""
        output_dir = os.path.join(temp_dir, "temporal_memory_output")
        # Create OpenAIVlModel instance
        api_key = os.getenv("OPENAI_API_KEY")
        vlm = OpenAIVlModel(api_key=api_key)

        temporal_memory = dimos_cluster.deploy(
            TemporalMemory,
            vlm=vlm,
            config=TemporalMemoryConfig(
                fps=1.0,  # Process 1 frame per second
                window_s=2.0,  # Analyze 2-second windows
                stride_s=2.0,  # New window every 2 seconds
                summary_interval_s=10.0,  # Update rolling summary every 10 seconds
                max_frames_per_window=3,  # Max 3 frames per window
                output_dir=output_dir,
            ),
        )
        yield temporal_memory
        try:
            temporal_memory.stop()
        except Exception as e:
            logger.warning(f"Failed to stop temporal_memory: {e}")

    @pytest.mark.asyncio
    async def test_temporal_memory_module_with_replay(
        self, dimos_cluster, video_module, temporal_memory, temp_dir
    ):
        """Test TemporalMemory module with TimedSensorReplay inputs."""
        # Connect streams
        temporal_memory.color_image.connect(video_module.video_out)

        # Start all modules
        video_module.start()
        temporal_memory.start()
        logger.info("All modules started, processing in background...")

        # Wait for frames to be processed with timeout
        timeout = 15.0  # 15 second timeout
        start_time = time.time()

        # Keep checking state while modules are running
        while (time.time() - start_time) < timeout:
            state = temporal_memory.get_state()
            if state["frame_count"] > 0:
                logger.info(
                    f"Frames processing - Frame count: {state['frame_count']}, "
                    f"Buffer size: {state['buffer_size']}, "
                    f"Entity count: {state['entity_count']}"
                )
                if state["frame_count"] >= 3:  # Wait for at least 3 frames
                    break
            await asyncio.sleep(0.5)
        else:
            # Timeout reached
            state = temporal_memory.get_state()
            logger.error(
                f"Timeout after {timeout}s - Frame count: {state['frame_count']}, "
                f"Buffer size: {state['buffer_size']}"
            )
            raise AssertionError(f"No frames processed within {timeout} seconds")

        await asyncio.sleep(3)  # Wait for more processing

        # Test get_state() RPC method
        mid_state = temporal_memory.get_state()
        logger.info(
            f"Mid-test state - Frame count: {mid_state['frame_count']}, "
            f"Entity count: {mid_state['entity_count']}, "
            f"Recent windows: {mid_state['recent_windows']}"
        )
        assert mid_state["frame_count"] >= state["frame_count"], (
            "Frame count should increase or stay same"
        )

        # Test query() RPC method
        answer = temporal_memory.query("What entities are currently visible?")
        logger.info(f"Query result: {answer[:200]}...")
        assert len(answer) > 0, "Query should return a non-empty answer"

        # Test get_entity_roster() RPC method
        entities = temporal_memory.get_entity_roster()
        logger.info(f"Entity roster has {len(entities)} entities")
        assert isinstance(entities, list), "Entity roster should be a list"

        # Test get_rolling_summary() RPC method
        summary = temporal_memory.get_rolling_summary()
        logger.info(f"Rolling summary: {summary[:200] if summary else 'empty'}...")
        assert isinstance(summary, str), "Rolling summary should be a string"

        final_state = temporal_memory.get_state()
        logger.info(
            f"Final state - Frame count: {final_state['frame_count']}, "
            f"Entity count: {final_state['entity_count']}, "
            f"Recent windows: {final_state['recent_windows']}"
        )

        video_module.stop()
        temporal_memory.stop()
        logger.info("Stopped modules")

        # Wait a bit for file operations to complete
        await asyncio.sleep(0.5)

        # Verify files were created - stop() already saved them
        output_dir = os.path.join(temp_dir, "temporal_memory_output")
        output_path = pathlib.Path(output_dir)
        assert output_path.exists(), f"Output directory should exist: {output_dir}"
        assert (output_path / "state.json").exists(), "state.json should exist"
        assert (output_path / "entities.json").exists(), "entities.json should exist"
        assert (output_path / "frames_index.jsonl").exists(), "frames_index.jsonl should exist"

        logger.info("All temporal memory module tests passed!")


if __name__ == "__main__":
    pytest.main(["-v", "-s", __file__])



#!/usr/bin/env python3
# Copyright 2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example usage of TemporalMemory module with a VLM.

This example demonstrates how to:
1. Deploy a camera module
2. Deploy TemporalMemory with the camera
3. Query the temporal memory about entities and events
"""

from pathlib import Path

from dotenv import load_dotenv

from dimos import core
from dimos.hardware.sensors.camera.module import CameraModule
from dimos.hardware.sensors.camera.webcam import Webcam

from .temporal_memory import TemporalMemoryConfig
from .temporal_memory_deploy import deploy

# Load environment variables
load_dotenv()


def example_usage() -> None:
    """Example of how to use TemporalMemory."""
    # Initialize variables to None for cleanup
    temporal_memory = None
    camera = None
    dimos = None

    try:
        # Create Dimos cluster
        dimos = core.start(1)
        # Deploy camera module
        camera = dimos.deploy(CameraModule, hardware=lambda: Webcam(camera_index=0))  # type: ignore[attr-defined]
        camera.start()

        # Deploy temporal memory using the deploy function
        output_dir = Path("./temporal_memory_output")
        temporal_memory = deploy(
            dimos,
            camera,
            vlm=None,  # Will auto-create OpenAIVlModel if None
            config=TemporalMemoryConfig(
                fps=1.0,  # Process 1 frame per second
                window_s=2.0,  # Analyze 2-second windows
                stride_s=2.0,  # New window every 2 seconds
                summary_interval_s=10.0,  # Update rolling summary every 10 seconds
                max_frames_per_window=3,  # Max 3 frames per window
                output_dir=output_dir,
            ),
        )

        print("TemporalMemory deployed and started!")
        print(f"Artifacts will be saved to: {output_dir}")

        # Let it run for a bit to build context
        print("Building temporal context... (wait ~15 seconds)")
        import time

        time.sleep(20)

        # Query the temporal memory
        questions = [
            "Are there any people in the scene?",
            "Describe the main activity happening now",
            "What has happened in the last few seconds?",
            "What entities are currently visible?",
        ]

        for question in questions:
            print(f"\nQuestion: {question}")
            answer = temporal_memory.query(question)
            print(f"Answer: {answer}")

        # Get current state
        state = temporal_memory.get_state()
        print("\n=== Current State ===")
        print(f"Entity count: {state['entity_count']}")
        print(f"Frame count: {state['frame_count']}")
        print(f"Rolling summary: {state['rolling_summary']}")
        print(f"Entities: {state['entities']}")

        # Get entity roster
        entities = temporal_memory.get_entity_roster()
        print("\n=== Entity Roster ===")
        for entity in entities:
            print(f"  {entity['id']}: {entity['descriptor']}")

        # Check graph database stats
        graph_stats = temporal_memory.get_graph_db_stats()
        print("\n=== Graph Database Stats ===")
        if "error" in graph_stats:
            print(f"Error: {graph_stats['error']}")
        else:
            print(f"Stats: {graph_stats['stats']}")
            print(f"\nEntities in DB ({len(graph_stats['entities'])}):")
            for entity in graph_stats["entities"]:
                print(f"  {entity['entity_id']} ({entity['entity_type']}): {entity['descriptor']}")
            print(f"\nRecent relations ({len(graph_stats['recent_relations'])}):")
            for rel in graph_stats["recent_relations"]:
                print(
                    f"  {rel['subject_id']} --{rel['relation_type']}--> {rel['object_id']} (confidence: {rel['confidence']:.2f})"
                )

        # Stop when done
        temporal_memory.stop()
        camera.stop()
        print("\nTemporalMemory stopped")

    finally:
        if temporal_memory is not None:
            temporal_memory.stop()
        if camera is not None:
            camera.stop()
        if dimos is not None:
            dimos.close_all()  # type: ignore[attr-defined]


if __name__ == "__main__":
    example_usage()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base agent module that wraps BaseAgent for DimOS module usage."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from dimos.protocol.skill.skill import SkillContainer

from langchain.chat_models.base import _SUPPORTED_PROVIDERS
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from rich.console import Console
from rich.table import Table
from rich.text import Text

from dimos.core import Module, rpc
from dimos.core.module import ModuleConfig
from dimos.protocol.pubsub import PubSub, lcm  # type: ignore[attr-defined]
from dimos.protocol.service import Service  # type: ignore[attr-defined]
from dimos.protocol.skill.skill import SkillContainer
from dimos.utils.generic import truncate_display_string
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


# Dynamically create ModelProvider enum from LangChain's supported providers
_providers = {provider.upper(): provider for provider in _SUPPORTED_PROVIDERS}
Provider = Enum("Provider", _providers, type=str)  # type: ignore[misc]


class Model(str, Enum):
    """Common model names across providers.

    Note: This is not exhaustive as model names change frequently.
    Based on langchain's _attempt_infer_model_provider patterns.
    """

    # OpenAI models (prefix: gpt-3, gpt-4, o1, o3)
    GPT_4O = "gpt-4o"
    GPT_4O_MINI = "gpt-4o-mini"
    GPT_4_TURBO = "gpt-4-turbo"
    GPT_4_TURBO_PREVIEW = "gpt-4-turbo-preview"
    GPT_4 = "gpt-4"
    GPT_35_TURBO = "gpt-3.5-turbo"
    GPT_35_TURBO_16K = "gpt-3.5-turbo-16k"
    O1_PREVIEW = "o1-preview"
    O1_MINI = "o1-mini"
    O3_MINI = "o3-mini"

    # Anthropic models (prefix: claude)
    CLAUDE_3_OPUS = "claude-3-opus-20240229"
    CLAUDE_3_SONNET = "claude-3-sonnet-20240229"
    CLAUDE_3_HAIKU = "claude-3-haiku-20240307"
    CLAUDE_35_SONNET = "claude-3-5-sonnet-20241022"
    CLAUDE_35_SONNET_LATEST = "claude-3-5-sonnet-latest"
    CLAUDE_3_7_SONNET = "claude-3-7-sonnet-20250219"

    # Google models (prefix: gemini)
    GEMINI_20_FLASH = "gemini-2.0-flash"
    GEMINI_15_PRO = "gemini-1.5-pro"
    GEMINI_15_FLASH = "gemini-1.5-flash"
    GEMINI_10_PRO = "gemini-1.0-pro"

    # Amazon Bedrock models (prefix: amazon)
    AMAZON_TITAN_EXPRESS = "amazon.titan-text-express-v1"
    AMAZON_TITAN_LITE = "amazon.titan-text-lite-v1"

    # Cohere models (prefix: command)
    COMMAND_R_PLUS = "command-r-plus"
    COMMAND_R = "command-r"
    COMMAND = "command"
    COMMAND_LIGHT = "command-light"

    # Fireworks models (prefix: accounts/fireworks)
    FIREWORKS_LLAMA_V3_70B = "accounts/fireworks/models/llama-v3-70b-instruct"
    FIREWORKS_MIXTRAL_8X7B = "accounts/fireworks/models/mixtral-8x7b-instruct"

    # Mistral models (prefix: mistral)
    MISTRAL_LARGE = "mistral-large"
    MISTRAL_MEDIUM = "mistral-medium"
    MISTRAL_SMALL = "mistral-small"
    MIXTRAL_8X7B = "mixtral-8x7b"
    MIXTRAL_8X22B = "mixtral-8x22b"
    MISTRAL_7B = "mistral-7b"

    # DeepSeek models (prefix: deepseek)
    DEEPSEEK_CHAT = "deepseek-chat"
    DEEPSEEK_CODER = "deepseek-coder"
    DEEPSEEK_R1_DISTILL_LLAMA_70B = "deepseek-r1-distill-llama-70b"

    # xAI models (prefix: grok)
    GROK_1 = "grok-1"
    GROK_2 = "grok-2"

    # Perplexity models (prefix: sonar)
    SONAR_SMALL_CHAT = "sonar-small-chat"
    SONAR_MEDIUM_CHAT = "sonar-medium-chat"
    SONAR_LARGE_CHAT = "sonar-large-chat"

    # Meta Llama models (various providers)
    LLAMA_3_70B = "llama-3-70b"
    LLAMA_3_8B = "llama-3-8b"
    LLAMA_31_70B = "llama-3.1-70b"
    LLAMA_31_8B = "llama-3.1-8b"
    LLAMA_33_70B = "llama-3.3-70b"
    LLAMA_2_70B = "llama-2-70b"
    LLAMA_2_13B = "llama-2-13b"
    LLAMA_2_7B = "llama-2-7b"


@dataclass
class AgentConfig(ModuleConfig):
    system_prompt: str | SystemMessage | None = None
    skills: SkillContainer | list[SkillContainer] | None = None

    # we can provide model/provvider enums or instantiated model_instance
    model: Model = Model.GPT_4O
    provider: Provider = Provider.OPENAI  # type: ignore[attr-defined]
    model_instance: BaseChatModel | None = None

    agent_transport: type[PubSub] = lcm.PickleLCM  # type: ignore[type-arg]
    agent_topic: Any = field(default_factory=lambda: lcm.Topic("/agent"))


AnyMessage = Union[SystemMessage, ToolMessage, AIMessage, HumanMessage]


class AgentSpec(Service[AgentConfig], Module, ABC):
    default_config: type[AgentConfig] = AgentConfig

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        Service.__init__(self, *args, **kwargs)
        Module.__init__(self, *args, **kwargs)

        if self.config.agent_transport:
            self.transport = self.config.agent_transport()

    def publish(self, msg: AnyMessage) -> None:
        if self.transport:
            self.transport.publish(self.config.agent_topic, msg)

    def start(self) -> None:
        super().start()

    def stop(self) -> None:
        if hasattr(self, "transport") and self.transport:
            self.transport.stop()  # type: ignore[attr-defined]
            self.transport = None  # type: ignore[assignment]
        super().stop()

    @rpc
    @abstractmethod
    def clear_history(self): ...  # type: ignore[no-untyped-def]

    @abstractmethod
    def append_history(self, *msgs: list[AIMessage | HumanMessage]): ...  # type: ignore[no-untyped-def]

    @abstractmethod
    def history(self) -> list[AnyMessage]: ...

    @rpc
    @abstractmethod
    def register_skills(
        self, container: "SkillContainer", run_implicit_name: str | None = None
    ) -> None: ...

    @rpc
    @abstractmethod
    def query(self, query: str): ...  # type: ignore[no-untyped-def]

    def __str__(self) -> str:
        console = Console(force_terminal=True, legacy_windows=False)
        table = Table(show_header=True)

        table.add_column("Message Type", style="cyan", no_wrap=True)
        table.add_column("Content")

        for message in self.history():
            if isinstance(message, HumanMessage):
                content = message.content
                if not isinstance(content, str):
                    content = "<Image>"

                table.add_row(Text("Human", style="green"), Text(content, style="green"))
            elif isinstance(message, AIMessage):
                if hasattr(message, "metadata") and message.metadata.get("state"):
                    table.add_row(
                        Text("State Summary", style="blue"),
                        Text(message.content, style="blue"),  # type: ignore[arg-type]
                    )
                else:
                    table.add_row(
                        Text("Agent", style="magenta"),
                        Text(message.content, style="magenta"),  # type: ignore[arg-type]
                    )

                for tool_call in message.tool_calls:
                    table.add_row(
                        "Tool Call",
                        Text(
                            f"{tool_call.get('name')}({tool_call.get('args')})",
                            style="bold magenta",
                        ),
                    )
            elif isinstance(message, ToolMessage):
                table.add_row(
                    "Tool Response", Text(f"{message.name}() -> {message.content}"), style="red"
                )
            elif isinstance(message, SystemMessage):
                table.add_row(
                    "System", Text(truncate_display_string(message.content, 800), style="yellow")
                )
            else:
                table.add_row("Unknown", str(message))

        # Render to string with title above
        with console.capture() as capture:
            console.print(Text(f"  Agent ({self._agent_id})", style="bold blue"))  # type: ignore[attr-defined]
            console.print(table)
        return capture.get().strip()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ANSI terminal renderer for module IO diagrams."""

from dimos.core import colors
from dimos.core.introspection.module.info import (
    ModuleInfo,
    ParamInfo,
    RpcInfo,
    SkillInfo,
    StreamInfo,
)


def render(info: ModuleInfo, color: bool = True) -> str:
    """Render module info as an ANSI terminal diagram.

    Args:
        info: ModuleInfo structure to render.
        color: Whether to include ANSI color codes.

    Returns:
        ASCII/Unicode diagram with optional ANSI colors.
    """
    # Color functions that become identity when color=False
    _green = colors.green if color else (lambda x: x)
    _blue = colors.blue if color else (lambda x: x)
    _yellow = colors.yellow if color else (lambda x: x)
    _cyan = colors.cyan if color else (lambda x: x)

    def _box(name: str) -> list[str]:
        return [
            "┌┴" + "─" * (len(name) + 1) + "┐",
            f"│ {name} │",
            "└┬" + "─" * (len(name) + 1) + "┘",
        ]

    def format_stream(stream: StreamInfo) -> str:
        return f"{_yellow(stream.name)}: {_green(stream.type_name)}"

    def format_param(param: ParamInfo) -> str:
        result = param.name
        if param.type_name:
            result += ": " + _green(param.type_name)
        if param.default:
            result += f" = {param.default}"
        return result

    def format_rpc(rpc: RpcInfo) -> str:
        params = ", ".join(format_param(p) for p in rpc.params)
        result = _blue(rpc.name) + f"({params})"
        if rpc.return_type:
            result += " -> " + _green(rpc.return_type)
        return result

    def format_skill(skill: SkillInfo) -> str:
        info_parts = []
        if skill.stream:
            info_parts.append(f"stream={skill.stream}")
        if skill.reducer:
            info_parts.append(f"reducer={skill.reducer}")
        if skill.output:
            info_parts.append(f"output={skill.output}")
        info = f" ({', '.join(info_parts)})" if info_parts else ""
        return _cyan(skill.name) + info

    # Build output
    lines = [
        *(f" ├─ {format_stream(s)}" for s in info.inputs),
        *_box(info.name),
        *(f" ├─ {format_stream(s)}" for s in info.outputs),
    ]

    if info.rpcs:
        lines.append(" │")
        for rpc in info.rpcs:
            lines.append(f" ├─ RPC {format_rpc(rpc)}")

    if info.skills:
        lines.append(" │")
        for skill in info.skills:
            lines.append(f" ├─ Skill {format_skill(skill)}")

    return "\n".join(lines)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for LocalModel and HuggingFaceModel base classes."""

from functools import cached_property

import torch

from dimos.models.base import HuggingFaceModel, LocalModel


class ConcreteLocalModel(LocalModel):
    """Concrete implementation for testing."""

    @cached_property
    def _model(self) -> str:
        return "loaded_model"


class ConcreteHuggingFaceModel(HuggingFaceModel):
    """Concrete implementation for testing."""

    @cached_property
    def _model(self) -> str:
        return f"hf_model:{self.model_name}"


def test_local_model_device_auto_detection() -> None:
    """Test that device is auto-detected based on CUDA availability."""
    model = ConcreteLocalModel()
    expected = "cuda" if torch.cuda.is_available() else "cpu"
    assert model.device == expected


def test_local_model_explicit_device() -> None:
    """Test that explicit device is respected."""
    model = ConcreteLocalModel(device="cpu")
    assert model.device == "cpu"


def test_local_model_default_dtype() -> None:
    """Test that default dtype is float32 for LocalModel."""
    model = ConcreteLocalModel()
    assert model.dtype == torch.float32


def test_local_model_explicit_dtype() -> None:
    """Test that explicit dtype is respected."""
    model = ConcreteLocalModel(dtype=torch.float16)
    assert model.dtype == torch.float16


def test_local_model_lazy_loading() -> None:
    """Test that model is lazily loaded."""
    model = ConcreteLocalModel()
    # Model not loaded yet
    assert "_model" not in model.__dict__
    # Access triggers loading
    _ = model._model
    # Now it's cached
    assert "_model" in model.__dict__
    assert model._model == "loaded_model"


def test_local_model_start_triggers_loading() -> None:
    """Test that start() triggers model loading."""
    model = ConcreteLocalModel()
    assert "_model" not in model.__dict__
    model.start()
    assert "_model" in model.__dict__


def test_huggingface_model_inherits_local_model() -> None:
    """Test that HuggingFaceModel inherits from LocalModel."""
    assert issubclass(HuggingFaceModel, LocalModel)


def test_huggingface_model_default_dtype() -> None:
    """Test that default dtype is float16 for HuggingFaceModel."""
    model = ConcreteHuggingFaceModel(model_name="test/model")
    assert model.dtype == torch.float16


def test_huggingface_model_name() -> None:
    """Test model_name property."""
    model = ConcreteHuggingFaceModel(model_name="microsoft/Florence-2-large")
    assert model.model_name == "microsoft/Florence-2-large"


def test_huggingface_model_trust_remote_code() -> None:
    """Test trust_remote_code defaults to True."""
    model = ConcreteHuggingFaceModel(model_name="test/model")
    assert model.config.trust_remote_code is True

    model2 = ConcreteHuggingFaceModel(model_name="test/model", trust_remote_code=False)
    assert model2.config.trust_remote_code is False


def test_huggingface_start_loads_model() -> None:
    """Test that start() loads model."""
    model = ConcreteHuggingFaceModel(model_name="test/model")
    assert "_model" not in model.__dict__
    model.start()
    assert "_model" in model.__dict__


def test_move_inputs_to_device() -> None:
    """Test _move_inputs_to_device helper."""
    model = ConcreteHuggingFaceModel(model_name="test/model", device="cpu")

    inputs = {
        "input_ids": torch.tensor([1, 2, 3]),
        "attention_mask": torch.tensor([1, 1, 1]),
        "pixel_values": torch.randn(1, 3, 224, 224),
        "labels": "not_a_tensor",
    }

    moved = model._move_inputs_to_device(inputs)

    assert moved["input_ids"].device.type == "cpu"
    assert moved["attention_mask"].device.type == "cpu"
    assert moved["pixel_values"].device.type == "cpu"
    assert moved["pixel_values"].dtype == torch.float16  # dtype applied
    assert moved["labels"] == "not_a_tensor"  # non-tensor unchanged



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from geometry_msgs.msg import Vector3  # type: ignore[attr-defined]
except ImportError:
    from dimos.msgs.geometry_msgs import Vector3

try:
    from geometry_msgs.msg import (  # type: ignore[attr-defined]
        Point,
        Pose,
        Quaternion,
        Twist,
    )
    from nav_msgs.msg import OccupancyGrid, Odometry  # type: ignore[attr-defined]
    from std_msgs.msg import Header  # type: ignore[attr-defined]
except ImportError:
    from dimos_lcm.geometry_msgs import (  # type: ignore[no-redef]
        Point,
        Pose,
        Quaternion,
        Twist,
    )
    from dimos_lcm.nav_msgs import OccupancyGrid, Odometry  # type: ignore[no-redef]
    from dimos_lcm.std_msgs import Header  # type: ignore[no-redef]

__all__ = [
    "Header",
    "OccupancyGrid",
    "Odometry",
    "Point",
    "Pose",
    "Quaternion",
    "Twist",
    "Vector3",
]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


def keepalive() -> None:
    try:
        # Keep the program running
        print("Press Ctrl+C to exit")
        print("-" * 60)
        while True:
            time.sleep(0.1)
    except KeyboardInterrupt:
        print("\nStopping pipeline")



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tiktoken

from dimos.agents_deprecated.tokenizer.base import AbstractTokenizer
from dimos.utils.logging_config import setup_logger


class OpenAITokenizer(AbstractTokenizer):
    def __init__(self, model_name: str = "gpt-4o", **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(**kwargs)

        # Initilize the tokenizer for the openai set of models
        self.model_name = model_name
        try:
            self.tokenizer = tiktoken.encoding_for_model(self.model_name)
        except Exception as e:
            raise ValueError(
                f"Failed to initialize tokenizer for model {self.model_name}. Error: {e!s}"
            )

    def tokenize_text(self, text: str):  # type: ignore[no-untyped-def]
        """
        Tokenize a text string using the openai tokenizer.
        """
        return self.tokenizer.encode(text)

    def detokenize_text(self, tokenized_text):  # type: ignore[no-untyped-def]
        """
        Detokenize a text string using the openai tokenizer.
        """
        try:
            return self.tokenizer.decode(tokenized_text, errors="ignore")
        except Exception as e:
            raise ValueError(f"Failed to detokenize text. Error: {e!s}")

    def token_count(self, text: str):  # type: ignore[no-untyped-def]
        """
        Gets the token count of a text string using the openai tokenizer.
        """
        return len(self.tokenize_text(text)) if text else 0

    @staticmethod
    def image_token_count(image_width, image_height, image_detail: str = "high"):  # type: ignore[no-untyped-def]
        """
        Calculate the number of tokens in an image. Low detail is 85 tokens, high detail is 170 tokens per 512x512 square.
        """
        logger = setup_logger()

        if image_detail == "low":
            return 85
        elif image_detail == "high":
            # Image dimensions
            logger.debug(f"Image Width: {image_width}, Image Height: {image_height}")
            if image_width is None or image_height is None:
                raise ValueError(
                    "Image width and height must be provided for high detail image token count calculation."
                )

            # Scale image to fit within 2048 x 2048
            max_dimension = max(image_width, image_height)
            if max_dimension > 2048:
                scale_factor = 2048 / max_dimension
                image_width = int(image_width * scale_factor)
                image_height = int(image_height * scale_factor)

            # Scale shortest side to 768px
            min_dimension = min(image_width, image_height)
            scale_factor = 768 / min_dimension
            image_width = int(image_width * scale_factor)
            image_height = int(image_height * scale_factor)

            # Calculate number of 512px squares
            num_squares = (image_width // 512) * (image_height // 512)
            return 170 * num_squares + 85
        else:
            raise ValueError("Detail specification of image is not 'low' or 'high'")



# Copyright (c) rxpy_backpressure
from typing import Any

from dimos.rxpy_backpressure.function_runner import thread_function_runner
from dimos.rxpy_backpressure.locks import BooleanLock, Lock
from dimos.rxpy_backpressure.observer import Observer


class DropBackPressureStrategy(Observer):
    def __init__(self, wrapped_observer: Observer, cache_size: int):
        self.wrapped_observer: Observer = wrapped_observer
        self.__function_runner = thread_function_runner
        self.__lock: Lock = BooleanLock()
        self.__cache_size: int | None = cache_size
        self.__message_cache: list = []
        self.__error_cache: list = []

    def on_next(self, message):
        if self.__lock.is_locked():
            self.__update_cache(self.__message_cache, message)
        else:
            self.__lock.lock()
            self.__function_runner(self, self.__on_next, message)

    @staticmethod
    def __on_next(self, message: any):
        self.wrapped_observer.on_next(message)
        if len(self.__message_cache) > 0:
            self.__function_runner(self, self.__on_next, self.__message_cache.pop(0))
        else:
            self.__lock.unlock()

    def on_error(self, error: any):
        if self.__lock.is_locked():
            self.__update_cache(self.__error_cache, error)
        else:
            self.__lock.lock()
            self.__function_runner(self, self.__on_error, error)

    @staticmethod
    def __on_error(self, error: any):
        self.wrapped_observer.on_error(error)
        if len(self.__error_cache) > 0:
            self.__function_runner(self, self.__on_error, self.__error_cache.pop(0))
        else:
            self.__lock.unlock()

    def __update_cache(self, cache: list, item: Any):
        if self.__cache_size is None or len(cache) < self.__cache_size:
            cache.append(item)
        else:
            cache.pop(0)
            cache.append(item)

    def on_completed(self):
        self.wrapped_observer.on_completed()

    def is_locked(self):
        return self.__lock.is_locked()


def wrap_observer_with_drop_strategy(observer: Observer, cache_size: int = 10) -> Observer:
    return DropBackPressureStrategy(observer, cache_size=cache_size)


def wrap_observer_with_buffer_strategy(observer: Observer) -> Observer:
    return DropBackPressureStrategy(observer, cache_size=None)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections.abc import Callable
import threading
from typing import Any, Protocol, overload


class Empty: ...


Args = tuple[list, dict[str, Any]]  # type: ignore[type-arg]


# module that we can inspect for RPCs
class RPCInspectable(Protocol):
    @property
    def rpcs(self) -> dict[str, Callable]: ...  # type: ignore[type-arg]


class RPCClient(Protocol):
    # if we don't provide callback, we don't get a return unsub f
    @overload
    def call(self, name: str, arguments: Args, cb: None) -> None: ...

    # if we provide callback, we do get return unsub f
    @overload
    def call(self, name: str, arguments: Args, cb: Callable[[Any], None]) -> Callable[[], Any]: ...

    def call(self, name: str, arguments: Args, cb: Callable | None) -> Callable[[], Any] | None: ...  # type: ignore[type-arg]

    # we expect to crash if we don't get a return value after 10 seconds
    # but callers can override this timeout for extra long functions
    def call_sync(
        self, name: str, arguments: Args, rpc_timeout: float | None = 120.0
    ) -> tuple[Any, Callable[[], None]]:
        if name == "start":
            rpc_timeout = 1200.0  # starting modules can take longer
        event = threading.Event()

        def receive_value(val) -> None:  # type: ignore[no-untyped-def]
            event.result = val  # type: ignore[attr-defined]  # attach to event
            event.set()

        unsub_fn = self.call(name, arguments, receive_value)
        if not event.wait(rpc_timeout):
            raise TimeoutError(f"RPC call to '{name}' timed out after {rpc_timeout} seconds")

        # Check if the result is an exception and raise it
        result = event.result  # type: ignore[attr-defined]
        if isinstance(result, BaseException):
            raise result

        return result, unsub_fn

    async def call_async(self, name: str, arguments: Args) -> Any:
        loop = asyncio.get_event_loop()
        future = loop.create_future()

        def receive_value(val) -> None:  # type: ignore[no-untyped-def]
            try:
                # Check if the value is an exception
                if isinstance(val, BaseException):
                    loop.call_soon_threadsafe(future.set_exception, val)
                else:
                    loop.call_soon_threadsafe(future.set_result, val)
            except Exception as e:
                loop.call_soon_threadsafe(future.set_exception, e)

        self.call(name, arguments, receive_value)

        return await future


class RPCServer(Protocol):
    def serve_rpc(self, f: Callable, name: str) -> Callable[[], None]: ...  # type: ignore[type-arg]

    def serve_module_rpc(self, module: RPCInspectable, name: str | None = None) -> None:
        for fname in module.rpcs.keys():
            if not name:
                name = module.__class__.__name__

            def override_f(*args, fname=fname, **kwargs):  # type: ignore[no-untyped-def]
                return getattr(module, fname)(*args, **kwargs)

            topic = name + "/" + fname
            self.serve_rpc(override_f, topic)


class RPCSpec(RPCServer, RPCClient): ...



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from dimos.perception.detection.type import ImageDetections2D


def test_from_ros_detection2d_array(get_moment_2d) -> None:
    moment = get_moment_2d()

    detections2d = moment["detections2d"]

    test_image = detections2d.image

    # Convert to ROS detection array
    ros_array = detections2d.to_ros_detection2d_array()

    # Convert back to ImageDetections2D
    recovered = ImageDetections2D.from_ros_detection2d_array(test_image, ros_array)

    # Verify we got the same number of detections
    assert len(recovered.detections) == len(detections2d.detections)

    # Verify the detection matches
    original_det = detections2d.detections[0]
    recovered_det = recovered.detections[0]

    # Check bbox is approximately the same (allow 1 pixel tolerance due to float conversion)
    for orig_val, rec_val in zip(original_det.bbox, recovered_det.bbox, strict=False):
        assert orig_val == pytest.approx(rec_val, abs=1.0)

    # Check other properties
    assert recovered_det.track_id == original_det.track_id
    assert recovered_det.class_id == original_det.class_id
    assert recovered_det.confidence == pytest.approx(original_det.confidence, abs=0.01)

    print("\nSuccessfully round-tripped detection through ROS format:")
    print(f"  Original bbox: {original_det.bbox}")
    print(f"  Recovered bbox: {recovered_det.bbox}")
    print(f"  Track ID: {recovered_det.track_id}")
    print(f"  Confidence: {recovered_det.confidence:.3f}")



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass, field
import time
from typing import TYPE_CHECKING, Any
import uuid

import cv2
from dimos_lcm.geometry_msgs import Pose
import numpy as np
import open3d as o3d  # type: ignore[import-untyped]

from dimos.msgs.geometry_msgs import PoseStamped, Quaternion, Transform, Vector3
from dimos.msgs.sensor_msgs import Image, PointCloud2
from dimos.msgs.std_msgs import Header
from dimos.msgs.vision_msgs import Detection3D as ROSDetection3D, Detection3DArray
from dimos.perception.detection.type.detection2d.seg import Detection2DSeg
from dimos.perception.detection.type.detection3d.base import Detection3D

if TYPE_CHECKING:
    from dimos_lcm.sensor_msgs import CameraInfo

    from dimos.perception.detection.type.detection2d import ImageDetections2D


@dataclass(kw_only=True)
class Object(Detection3D):
    """3D object detection combining bounding box and pointcloud representations.

    Represents a detected object in 3D space with support for accumulating
    multiple detections over time.
    """

    object_id: str = field(default_factory=lambda: uuid.uuid4().hex[:8])
    center: Vector3
    size: Vector3
    pose: PoseStamped
    pointcloud: PointCloud2
    camera_transform: Transform | None = None
    mask: np.ndarray[Any, np.dtype[np.uint8]] | None = None
    detections_count: int = 1

    def update_object(self, other: Object) -> None:
        """Update this object with data from another detection.

        Accumulates pointclouds by transforming the new pointcloud to world frame
        and adding it to the existing pointcloud. Updates center and camera_transform,
        and increments the detections_count.

        Args:
            other: Another Object instance with newer detection data.
        """
        # Accumulate pointclouds if transforms are available
        if other.camera_transform is not None:
            # Transform new pointcloud to world frame and add to existing
            # transformed_pc = other.pointcloud.transform(other.camera_transform)
            # self.pointcloud = self.pointcloud + transformed_pc

            # Recompute center from accumulated pointcloud
            self.pointcloud = other.pointcloud
            pc_center = other.pointcloud.center
            self.center = Vector3(pc_center.x, pc_center.y, pc_center.z)
        else:
            # No transform available, just replace
            self.pointcloud = other.pointcloud
            self.center = other.center

        self.camera_transform = other.camera_transform
        self.size = other.size
        self.pose = other.pose
        self.track_id = other.track_id
        self.mask = other.mask
        self.name = other.name
        self.bbox = other.bbox
        self.confidence = other.confidence
        self.class_id = other.class_id
        self.ts = other.ts
        self.frame_id = other.frame_id
        self.image = other.image
        self.detections_count += 1

    def get_oriented_bounding_box(self) -> Any:
        """Get oriented bounding box of the pointcloud."""
        return self.pointcloud.get_oriented_bounding_box()

    def scene_entity_label(self) -> str:
        """Get label for scene visualization."""
        if self.detections_count > 1:
            return f"{self.name} ({self.detections_count})"
        return f"{self.track_id}/{self.name} ({self.confidence:.0%})"

    def to_detection3d_msg(self) -> ROSDetection3D:
        """Convert to ROS Detection3D message."""
        obb = self.get_oriented_bounding_box()  # type: ignore[no-untyped-call]
        orientation = Quaternion.from_rotation_matrix(obb.R)

        msg = ROSDetection3D()
        msg.header = Header(self.ts, self.frame_id)
        msg.id = str(self.track_id)
        msg.bbox.center = Pose(
            position=Vector3(obb.center[0], obb.center[1], obb.center[2]),
            orientation=orientation,
        )
        msg.bbox.size = Vector3(obb.extent[0], obb.extent[1], obb.extent[2])

        return msg

    def agent_encode(self) -> dict[str, Any]:
        """Encode for agent consumption."""
        return {
            "id": self.track_id,
            "name": self.name,
            "detections": self.detections_count,
            "last_seen": f"{round(time.time() - self.ts)}s ago",
        }

    def to_dict(self) -> dict[str, Any]:
        """Convert object to dictionary with all relevant data."""
        return {
            "object_id": self.object_id,
            "track_id": self.track_id,
            "class_id": self.class_id,
            "name": self.name,
            "mask": self.mask,
            "pointcloud": self.pointcloud.as_numpy(),
            "image": self.image.as_numpy() if self.image else None,
        }

    @classmethod
    def from_2d_to_list(
        cls,
        detections_2d: ImageDetections2D[Detection2DSeg],
        color_image: Image,
        depth_image: Image,
        camera_info: CameraInfo,
        camera_transform: Transform | None = None,
        depth_scale: float = 1.0,
        depth_trunc: float = 10.0,
        statistical_nb_neighbors: int = 10,
        statistical_std_ratio: float = 0.5,
        voxel_downsample: float = 0.005,
        mask_erode_pixels: int = 3,
    ) -> list[Object]:
        """Create 3D Objects from 2D detections and RGBD images.

        Uses Open3D's optimized RGBD projection for efficient processing.

        Args:
            detections_2d: 2D detections with segmentation masks
            color_image: RGB color image
            depth_image: Depth image (in meters if depth_scale=1.0)
            camera_info: Camera intrinsics
            camera_transform: Optional transform from camera frame to world frame.
                If provided, pointclouds will be transformed to world frame.
            depth_scale: Scale factor for depth (1.0 for meters, 1000.0 for mm)
            depth_trunc: Maximum depth value in meters
            statistical_nb_neighbors: Neighbors for statistical outlier removal
            statistical_std_ratio: Std ratio for statistical outlier removal
            voxel_downsample: Voxel size (meters) for downsampling before filtering. Set <= 0 to skip.
            mask_erode_pixels: Number of pixels to erode the mask by to remove
                              noisy depth edge points. Set to 0 to disable.

        Returns:
            List of Object instances with pointclouds
        """
        color_cv = color_image.to_opencv()
        if color_cv.ndim == 3 and color_cv.shape[2] == 3:
            color_cv = cv2.cvtColor(color_cv, cv2.COLOR_BGR2RGB)

        depth_cv = depth_image.to_opencv()
        h, w = depth_cv.shape[:2]

        # Build Open3D camera intrinsics
        fx, fy = camera_info.K[0], camera_info.K[4]
        cx, cy = camera_info.K[2], camera_info.K[5]
        intrinsic_o3d = o3d.camera.PinholeCameraIntrinsic(w, h, fx, fy, cx, cy)

        objects: list[Object] = []

        for det in detections_2d.detections:
            if isinstance(det, Detection2DSeg):
                mask = det.mask
                store_mask = det.mask
            else:
                mask = np.zeros((h, w), dtype=np.uint8)
                x1, y1, x2, y2 = map(int, det.bbox)
                x1, y1 = max(0, x1), max(0, y1)
                x2, y2 = min(w, x2), min(h, y2)
                mask[y1:y2, x1:x2] = 255
                store_mask = mask

            if mask_erode_pixels > 0:
                mask_uint8 = mask.astype(np.uint8)
                if mask_uint8.max() == 1:
                    mask_uint8 = mask_uint8 * 255  # type: ignore[assignment]
                kernel_size = 2 * mask_erode_pixels + 1
                erode_kernel = cv2.getStructuringElement(
                    cv2.MORPH_ELLIPSE, (kernel_size, kernel_size)
                )
                mask = cv2.erode(mask_uint8, erode_kernel)  # type: ignore[assignment]

            depth_masked = depth_cv.copy()
            depth_masked[mask == 0] = 0

            rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
                o3d.geometry.Image(color_cv.astype(np.uint8)),
                o3d.geometry.Image(depth_masked.astype(np.float32)),
                depth_scale=depth_scale,
                depth_trunc=depth_trunc,
                convert_rgb_to_intensity=False,
            )
            pcd = o3d.geometry.PointCloud.create_from_rgbd_image(rgbd, intrinsic_o3d)

            pc0 = PointCloud2(
                pcd,
                frame_id=depth_image.frame_id,
                ts=depth_image.ts,
            ).voxel_downsample(voxel_downsample)

            pcd_filtered, _ = pc0.pointcloud.remove_statistical_outlier(
                nb_neighbors=statistical_nb_neighbors,
                std_ratio=statistical_std_ratio,
            )

            if len(pcd_filtered.points) < 10:
                continue

            pc = PointCloud2(
                pcd_filtered,
                frame_id=depth_image.frame_id,
                ts=depth_image.ts,
            )

            # Transform pointcloud to world frame if camera_transform is provided
            if camera_transform is not None:
                pc = pc.transform(camera_transform)
                frame_id = camera_transform.frame_id
            else:
                frame_id = depth_image.frame_id

            # Compute center from pointcloud
            obb = pc.pointcloud.get_oriented_bounding_box()
            center = Vector3(obb.center[0], obb.center[1], obb.center[2])
            size = Vector3(obb.extent[0], obb.extent[1], obb.extent[2])
            orientation = Quaternion.from_rotation_matrix(obb.R)
            pose = PoseStamped(
                ts=det.ts,
                frame_id=frame_id,
                position=center,
                orientation=orientation,
            )

            objects.append(
                cls(
                    bbox=det.bbox,
                    track_id=det.track_id,
                    class_id=det.class_id,
                    confidence=det.confidence,
                    name=det.name,
                    ts=det.ts,
                    image=det.image,
                    frame_id=frame_id,
                    pointcloud=pc,
                    center=center,
                    size=size,
                    pose=pose,
                    camera_transform=camera_transform,
                    mask=store_mask,
                )
            )

        return objects


def aggregate_pointclouds(objects: list[Object]) -> PointCloud2:
    """Aggregate all object pointclouds into a single colored pointcloud.

    Each object's points are colored based on its track_id.

    Args:
        objects: List of Object instances with pointclouds

    Returns:
        Combined PointCloud2 with all points colored by object (empty if no points).
    """
    if not objects:
        return PointCloud2(pointcloud=o3d.geometry.PointCloud(), frame_id="", ts=0.0)

    all_points = []
    all_colors = []

    for _i, obj in enumerate(objects):
        points, colors = obj.pointcloud.as_numpy()
        if len(points) == 0:
            continue

        try:
            seed = int(obj.object_id, 16)
        except (ValueError, TypeError):
            seed = abs(hash(obj.object_id))
        np.random.seed(abs(seed) % (2**32 - 1))
        track_color = np.random.randint(50, 255, 3) / 255.0

        if colors is not None:
            blended = np.clip(0.6 * colors + 0.4 * track_color, 0.0, 1.0)
        else:
            blended = np.tile(track_color, (len(points), 1))

        all_points.append(points)
        all_colors.append(blended)

    if not all_points:
        return PointCloud2(
            pointcloud=o3d.geometry.PointCloud(), frame_id=objects[0].frame_id, ts=objects[0].ts
        )

    combined_points = np.vstack(all_points)
    combined_colors = np.vstack(all_colors)

    pc = PointCloud2.from_numpy(
        combined_points,
        frame_id=objects[0].frame_id,
        timestamp=objects[0].ts,
    )
    pcd = pc.pointcloud
    pcd.colors = o3d.utility.Vector3dVector(combined_colors)
    pc.pointcloud = pcd

    return pc


def to_detection3d_array(objects: list[Object]) -> Detection3DArray:
    """Convert a list of Objects to a ROS Detection3DArray message.

    Args:
        objects: List of Object instances

    Returns:
        Detection3DArray ROS message
    """
    array = Detection3DArray()

    if objects:
        array.header = Header(objects[0].ts, objects[0].frame_id)

    for obj in objects:
        array.detections.append(obj.to_detection3d_msg())

    return array



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from typing import TypeAlias

from dimos_lcm.sensor_msgs import JointState as LCMJointState

try:
    from sensor_msgs.msg import JointState as ROSJointState  # type: ignore[attr-defined]
except ImportError:
    ROSJointState = None  # type: ignore[assignment, misc]

from plum import dispatch

from dimos.types.timestamped import Timestamped

# Types that can be converted to/from JointState
JointStateConvertable: TypeAlias = dict[str, list[str] | list[float]] | LCMJointState


def sec_nsec(ts):  # type: ignore[no-untyped-def]
    s = int(ts)
    return [s, int((ts - s) * 1_000_000_000)]


class JointState(Timestamped):
    msg_name = "sensor_msgs.JointState"
    ts: float
    frame_id: str
    name: list[str]
    position: list[float]
    velocity: list[float]
    effort: list[float]

    @dispatch
    def __init__(
        self,
        ts: float = 0.0,
        frame_id: str = "",
        name: list[str] | None = None,
        position: list[float] | None = None,
        velocity: list[float] | None = None,
        effort: list[float] | None = None,
    ) -> None:
        """Initialize a JointState message.

        Args:
            ts: Timestamp in seconds
            frame_id: Frame ID for the message
            name: List of joint names
            position: List of joint positions (rad or m)
            velocity: List of joint velocities (rad/s or m/s)
            effort: List of joint efforts (Nm or N)
        """
        self.ts = ts if ts != 0 else time.time()
        self.frame_id = frame_id
        self.name = name if name is not None else []
        self.position = position if position is not None else []
        self.velocity = velocity if velocity is not None else []
        self.effort = effort if effort is not None else []

    @dispatch  # type: ignore[no-redef]
    def __init__(self, joint_dict: dict[str, list[str] | list[float]]) -> None:
        """Initialize from a dictionary."""
        self.ts = joint_dict.get("ts", time.time())
        self.frame_id = joint_dict.get("frame_id", "")
        self.name = list(joint_dict.get("name", []))
        self.position = list(joint_dict.get("position", []))
        self.velocity = list(joint_dict.get("velocity", []))
        self.effort = list(joint_dict.get("effort", []))

    @dispatch  # type: ignore[no-redef]
    def __init__(self, joint: JointState) -> None:
        """Initialize from another JointState (copy constructor)."""
        self.ts = joint.ts
        self.frame_id = joint.frame_id
        self.name = list(joint.name)
        self.position = list(joint.position)
        self.velocity = list(joint.velocity)
        self.effort = list(joint.effort)

    @dispatch  # type: ignore[no-redef]
    def __init__(self, lcm_joint: LCMJointState) -> None:
        """Initialize from an LCM JointState message."""
        self.ts = lcm_joint.header.stamp.sec + (lcm_joint.header.stamp.nsec / 1_000_000_000)
        self.frame_id = lcm_joint.header.frame_id
        self.name = list(lcm_joint.name) if lcm_joint.name else []
        self.position = list(lcm_joint.position) if lcm_joint.position else []
        self.velocity = list(lcm_joint.velocity) if lcm_joint.velocity else []
        self.effort = list(lcm_joint.effort) if lcm_joint.effort else []

    def lcm_encode(self) -> bytes:
        lcm_msg = LCMJointState()
        [lcm_msg.header.stamp.sec, lcm_msg.header.stamp.nsec] = sec_nsec(self.ts)  # type: ignore[no-untyped-call]
        lcm_msg.header.frame_id = self.frame_id
        lcm_msg.name_length = len(self.name)
        lcm_msg.name = self.name
        lcm_msg.position_length = len(self.position)
        lcm_msg.position = self.position
        lcm_msg.velocity_length = len(self.velocity)
        lcm_msg.velocity = self.velocity
        lcm_msg.effort_length = len(self.effort)
        lcm_msg.effort = self.effort
        return lcm_msg.lcm_encode()  # type: ignore[no-any-return]

    @classmethod
    def lcm_decode(cls, data: bytes) -> JointState:
        lcm_msg = LCMJointState.lcm_decode(data)
        return cls(
            ts=lcm_msg.header.stamp.sec + (lcm_msg.header.stamp.nsec / 1_000_000_000),
            frame_id=lcm_msg.header.frame_id,
            name=list(lcm_msg.name) if lcm_msg.name else [],
            position=list(lcm_msg.position) if lcm_msg.position else [],
            velocity=list(lcm_msg.velocity) if lcm_msg.velocity else [],
            effort=list(lcm_msg.effort) if lcm_msg.effort else [],
        )

    def __str__(self) -> str:
        return f"JointState({len(self.name)} joints, frame_id='{self.frame_id}')"

    def __repr__(self) -> str:
        return (
            f"JointState(ts={self.ts}, frame_id='{self.frame_id}', "
            f"name={self.name}, position={self.position}, "
            f"velocity={self.velocity}, effort={self.effort})"
        )

    def __eq__(self, other) -> bool:  # type: ignore[no-untyped-def]
        """Check if two JointState messages are equal."""
        if not isinstance(other, JointState):
            return False
        return (
            self.name == other.name
            and self.position == other.position
            and self.velocity == other.velocity
            and self.effort == other.effort
            and self.frame_id == other.frame_id
        )

    @classmethod
    def from_ros_msg(cls, ros_msg: ROSJointState) -> JointState:
        """Create a JointState from a ROS sensor_msgs/JointState message.

        Args:
            ros_msg: ROS JointState message

        Returns:
            JointState instance
        """
        # Convert timestamp from ROS header
        ts = ros_msg.header.stamp.sec + (ros_msg.header.stamp.nanosec / 1_000_000_000)

        return cls(
            ts=ts,
            frame_id=ros_msg.header.frame_id,
            name=list(ros_msg.name),
            position=list(ros_msg.position),
            velocity=list(ros_msg.velocity),
            effort=list(ros_msg.effort),
        )

    def to_ros_msg(self) -> ROSJointState:
        """Convert to a ROS sensor_msgs/JointState message.

        Returns:
            ROS JointState message
        """
        ros_msg = ROSJointState()  # type: ignore[no-untyped-call]

        # Set header
        ros_msg.header.frame_id = self.frame_id
        ros_msg.header.stamp.sec = int(self.ts)
        ros_msg.header.stamp.nanosec = int((self.ts - int(self.ts)) * 1_000_000_000)

        # Set joint data
        ros_msg.name = self.name
        ros_msg.position = self.position
        ros_msg.velocity = self.velocity
        ros_msg.effort = self.effort

        return ros_msg



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

from dimos.core import In, Module, Out, rpc


class Counter(Module):
    current_count: int = 0

    count_stream: Out[int]

    def __init__(self) -> None:
        super().__init__()
        self.current_count = 0

    @rpc
    def increment(self):
        """Increment the counter and publish the new value."""
        self.current_count += 1
        self.count_stream.publish(self.current_count)
        return self.current_count


class CounterValidator(Module):
    """Calls counter.increment() as fast as possible and validates no numbers are skipped."""

    count_in: In[int]

    def __init__(self, increment_func) -> None:
        super().__init__()
        self.increment_func = increment_func
        self.last_seen = 0
        self.missing_numbers = []
        self.running = False
        self.call_thread = None
        self.call_count = 0
        self.total_latency = 0.0
        self.call_start_time = None
        self.waiting_for_response = False

    @rpc
    def start(self) -> None:
        """Start the validator."""
        self.count_in.subscribe(self._on_count_received)
        self.running = True
        self.call_thread = threading.Thread(target=self._call_loop)
        self.call_thread.start()

    @rpc
    def stop(self) -> None:
        """Stop the validator."""
        self.running = False
        if self.call_thread:
            self.call_thread.join()

    def _on_count_received(self, count: int) -> None:
        """Check if we received all numbers in sequence and trigger next call."""
        # Calculate round trip time
        if self.call_start_time:
            latency = time.time() - self.call_start_time
            self.total_latency += latency

        if count != self.last_seen + 1:
            for missing in range(self.last_seen + 1, count):
                self.missing_numbers.append(missing)
                print(f"[VALIDATOR] Missing number detected: {missing}")
        self.last_seen = count

        # Signal that we can make the next call
        self.waiting_for_response = False

    def _call_loop(self) -> None:
        """Call increment only after receiving response from previous call."""
        while self.running:
            if not self.waiting_for_response:
                try:
                    self.waiting_for_response = True
                    self.call_start_time = time.time()
                    result = self.increment_func()
                    call_time = time.time() - self.call_start_time
                    self.call_count += 1
                    if self.call_count % 100 == 0:
                        avg_latency = (
                            self.total_latency / self.call_count if self.call_count > 0 else 0
                        )
                        print(
                            f"[VALIDATOR] Made {self.call_count} calls, last result: {result}, RPC call time: {call_time * 1000:.2f}ms, avg RTT: {avg_latency * 1000:.2f}ms"
                        )
                except Exception as e:
                    print(f"[VALIDATOR] Error calling increment: {e}")
                    self.waiting_for_response = False
                    time.sleep(0.001)  # Small delay on error
            else:
                # Don't sleep - busy wait for maximum speed
                pass

    @rpc
    def get_stats(self):
        """Get validation statistics."""
        avg_latency = self.total_latency / self.call_count if self.call_count > 0 else 0
        return {
            "call_count": self.call_count,
            "last_seen": self.last_seen,
            "missing_count": len(self.missing_numbers),
            "missing_numbers": self.missing_numbers[:10] if self.missing_numbers else [],
            "avg_rtt_ms": avg_latency * 1000,
            "calls_per_sec": self.call_count / 10.0 if self.call_count > 0 else 0,
        }


if __name__ == "__main__":
    import dimos.core as core
    from dimos.core import pLCMTransport

    # Start dimos with 2 workers
    client = core.start(2)

    # Deploy counter module
    counter = client.deploy(Counter)
    counter.count_stream.transport = pLCMTransport("/counter_stream")

    # Deploy validator module with increment function
    validator = client.deploy(CounterValidator, counter.increment)
    validator.count_in.transport = pLCMTransport("/counter_stream")

    # Connect validator to counter's output
    validator.count_in.connect(counter.count_stream)

    # Start modules
    validator.start()

    print("[MAIN] Counter and validator started. Running for 10 seconds...")

    # Test direct RPC speed for comparison
    print("\n[MAIN] Testing direct RPC call speed for 1 second...")
    start = time.time()
    direct_count = 0
    while time.time() - start < 1.0:
        counter.increment()
        direct_count += 1
    print(f"[MAIN] Direct RPC calls per second: {direct_count}")

    # Run for 10 seconds
    time.sleep(10)

    # Get stats before stopping
    stats = validator.get_stats()
    print("\n[MAIN] Final statistics:")
    print(f"  - Total calls made: {stats['call_count']}")
    print(f"  - Last number seen: {stats['last_seen']}")
    print(f"  - Missing numbers: {stats['missing_count']}")
    print(f"  - Average RTT: {stats['avg_rtt_ms']:.2f}ms")
    print(f"  - Calls per second: {stats['calls_per_sec']:.1f}")
    if stats["missing_numbers"]:
        print(f"  - First missing numbers: {stats['missing_numbers']}")

    # Stop modules
    validator.stop()

    # Shutdown dimos
    client.shutdown()

    print("[MAIN] Test complete.")



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2025-2026 Dimensional Inc.

"""Int32 message type."""

from typing import ClassVar

from dimos_lcm.std_msgs import Int8 as LCMInt8

try:
    from std_msgs.msg import Int8 as ROSInt8  # type: ignore[attr-defined]
except ImportError:
    ROSInt8 = None  # type: ignore[assignment, misc]


class Int8(LCMInt8):  # type: ignore[misc]
    """ROS-compatible Int32 message."""

    msg_name: ClassVar[str] = "std_msgs.Int8"

    def __init__(self, data: int = 0) -> None:
        """Initialize Int8 with data value."""
        self.data = data

    @classmethod
    def from_ros_msg(cls, ros_msg: ROSInt8) -> "Int8":
        """Create a Bool from a ROS std_msgs/Bool message.

        Args:
            ros_msg: ROS Int8 message

        Returns:
            Int8 instance
        """
        return cls(data=ros_msg.data)

    def to_ros_msg(self) -> ROSInt8:
        """Convert to a ROS std_msgs/Bool message.

        Returns:
            ROS Int8 message
        """
        if ROSInt8 is None:
            raise ImportError("ROS std_msgs not available")
        ros_msg = ROSInt8()  # type: ignore[no-untyped-call]
        ros_msg.data = self.data
        return ros_msg



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from typing_extensions import TypeVar

from dimos.perception.detection.type.detection2d.base import Detection2D
from dimos.perception.detection.type.detection2d.bbox import Detection2DBBox
from dimos.perception.detection.type.detection2d.person import Detection2DPerson
from dimos.perception.detection.type.detection2d.seg import Detection2DSeg
from dimos.perception.detection.type.imageDetections import ImageDetections

if TYPE_CHECKING:
    from ultralytics.engine.results import Results

    from dimos.msgs.sensor_msgs import Image
    from dimos.msgs.vision_msgs import Detection2DArray

T2D = TypeVar("T2D", bound=Detection2D, default=Detection2DBBox)


class ImageDetections2D(ImageDetections[T2D], Generic[T2D]):
    @classmethod
    def from_ros_detection2d_array(
        cls,
        image: Image,
        ros_detections: Detection2DArray,
        **kwargs: Any,
    ) -> ImageDetections2D[Detection2DBBox]:
        """Convert from ROS Detection2DArray message to ImageDetections2D object."""
        detections: list[Detection2DBBox] = []
        for ros_det in ros_detections.detections:
            detection = Detection2DBBox.from_ros_detection2d(ros_det, image=image, **kwargs)
            if detection.is_valid():
                detections.append(detection)

        return ImageDetections2D(image=image, detections=detections)

    @classmethod
    def from_ultralytics_result(
        cls,
        image: Image,
        results: list[Results],
    ) -> ImageDetections2D[Detection2DBBox]:
        """Create ImageDetections2D from ultralytics Results.

        Dispatches to appropriate Detection2D subclass based on result type:
        - If masks present: creates Detection2DSeg
        - If keypoints present: creates Detection2DPerson
        - Otherwise: creates Detection2DBBox

        Args:
            image: Source image
            results: List of ultralytics Results objects

        Returns:
            ImageDetections2D containing appropriate detection types
        """

        detections: list[Detection2DBBox] = []
        for result in results:
            if result.boxes is None:
                continue

            num_detections = len(result.boxes.xyxy)
            for i in range(num_detections):
                detection: Detection2DBBox
                if result.masks is not None:
                    # Segmentation detection with mask
                    detection = Detection2DSeg.from_ultralytics_result(result, i, image)
                elif result.keypoints is not None:
                    # Pose detection with keypoints
                    detection = Detection2DPerson.from_ultralytics_result(result, i, image)
                else:
                    # Regular bbox detection
                    detection = Detection2DBBox.from_ultralytics_result(result, i, image)
                if detection.is_valid():
                    detections.append(detection)

        return ImageDetections2D(image=image, detections=detections)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dimos.mapping.types import LatLon


def test_set_gps_travel_points(create_gps_nav_agent, gps_nav_skill_container, mocker) -> None:
    gps_nav_skill_container._latest_location = LatLon(lat=37.782654, lon=-122.413273)
    gps_nav_skill_container._set_gps_travel_goal_points = mocker.Mock()
    agent = create_gps_nav_agent(fixture="test_set_gps_travel_points.json")

    agent.query("go to lat: 37.782654, lon: -122.413273")

    gps_nav_skill_container._set_gps_travel_goal_points.assert_called_once_with(
        [LatLon(lat=37.782654, lon=-122.413273)]
    )
    gps_nav_skill_container.gps_goal.publish.assert_called_once_with(
        [LatLon(lat=37.782654, lon=-122.413273)]
    )


def test_set_gps_travel_points_multiple(
    create_gps_nav_agent, gps_nav_skill_container, mocker
) -> None:
    gps_nav_skill_container._latest_location = LatLon(lat=37.782654, lon=-122.413273)
    gps_nav_skill_container._set_gps_travel_goal_points = mocker.Mock()
    agent = create_gps_nav_agent(fixture="test_set_gps_travel_points_multiple.json")

    agent.query(
        "go to lat: 37.782654, lon: -122.413273, then 37.782660,-122.413260, and then 37.782670,-122.413270"
    )

    gps_nav_skill_container._set_gps_travel_goal_points.assert_called_once_with(
        [
            LatLon(lat=37.782654, lon=-122.413273),
            LatLon(lat=37.782660, lon=-122.413260),
            LatLon(lat=37.782670, lon=-122.413270),
        ]
    )
    gps_nav_skill_container.gps_goal.publish.assert_called_once_with(
        [
            LatLon(lat=37.782654, lon=-122.413273),
            LatLon(lat=37.782660, lon=-122.413260),
            LatLon(lat=37.782670, lon=-122.413270),
        ]
    )
